import logging
from pathlib import Path

from housekeeper.store.models import File

from cg.apps.gt import GenotypeAPI
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants.constants import FileExtensions, FileFormat, PrepCategory, Workflow
from cg.constants.housekeeper_tags import GenotypeAnalysisTag, HkAnalysisMetricsTag
from cg.constants.nf_analysis import RAREDISEASE_PREDICTED_SEX_METRIC
from cg.constants.subject import Sex
from cg.io.controller import ReadFile
from cg.models.deliverables.metric_deliverables import MetricsBase
from cg.models.mip.mip_metrics_deliverables import MIPMetricsDeliverables
from cg.store.models import Analysis, Case, Sample

LOG = logging.getLogger(__name__)


class UploadGenotypesAPI(object):
    """Genotype upload API."""

    def __init__(
        self,
        hk_api: HousekeeperAPI,
        gt_api: GenotypeAPI,
    ):
        LOG.info("Initializing UploadGenotypesAPI")
        self.hk = hk_api
        self.gt = gt_api

    def get_genotype_data(self, analysis: Analysis) -> dict[dict[str, dict[str, str]]]:
        """Return data about an analysis to load genotypes.

        Returns: dict on form

        {
            "bcf": path_to_bcf,
            "samples_sex": [
                "sample_id: {
                    "pedigree": "male",
                    "analysis": "male"
                    }
            ]
        }

        """
        case_id = analysis.case.internal_id
        LOG.info(f"Fetching upload genotype data for {case_id}")
        hk_bcf: File = self._get_genotype_file(case_id=case_id)
        genotype_load_config: dict = {"bcf": hk_bcf.full_path}
        if analysis.workflow in [Workflow.BALSAMIC, Workflow.BALSAMIC_UMI]:
            genotype_load_config["samples_sex"] = self._get_samples_sex_balsamic(case=analysis.case)
        elif analysis.workflow == Workflow.MIP_DNA:
            genotype_load_config["samples_sex"] = self._get_samples_sex_mip_dna(case=analysis.case)
        elif analysis.workflow == Workflow.RAREDISEASE:
            genotype_load_config["samples_sex"] = self._get_samples_sex_raredisease(
                case=analysis.case
            )
        else:
            raise ValueError(f"Workflow {analysis.workflow} does not support Genotype upload")
        return genotype_load_config

    def upload(self, genotype_load_config: dict, replace: bool = False):
        """Upload a genotype config for a case."""
        self.gt.upload(
            str(genotype_load_config["bcf"]), genotype_load_config["samples_sex"], force=replace
        )

    @staticmethod
    def is_suitable_for_genotype_upload(case: Case) -> bool:
        """Returns True if there are any non-tumor WGS samples in the case."""
        samples: list[Sample] = case.samples
        return any(
            (not sample.is_tumour and PrepCategory.WHOLE_GENOME_SEQUENCING == sample.prep_category)
            for sample in samples
        )

    def _get_genotype_file(self, case_id: str) -> File:
        "Returns latest genotype file in housekeeper for given case, raises FileNotFoundError if not found."
        LOG.debug("Get Genotype files from Housekeeper")
        tags: set[str] = {GenotypeAnalysisTag.GENOTYPE}
        hk_genotype_files: list[File] = self.hk.get_files_from_latest_version(
            bundle_name=case_id, tags=tags
        )
        hk_genotype: File = self._get_single_genotype_file(hk_genotype_files)
        if not hk_genotype:
            raise FileNotFoundError(f"Genotype file not found for {case_id}")
        LOG.debug(f"Found genotype file {hk_genotype.full_path}")
        return hk_genotype

    def _get_samples_sex_balsamic(self, case: Case) -> dict[str, dict[str, str]]:
        """Return sex information from StatusDB and for analysis."""
        samples_sex: dict[str, dict[str, str]] = {}
        for sample in case.samples:
            if sample.is_tumour:
                continue
            sample_id: str = sample.internal_id
            samples_sex[sample_id] = {
                "pedigree": sample.sex,
                "analysis": Sex.UNKNOWN,
            }
        return samples_sex

    def _get_samples_sex_mip_dna(self, case: Case) -> dict[str, dict[str, str]]:
        """Return sex information from StatusDB and from analysis prediction"""
        qc_metrics_file: Path = self._get_qcmetrics_file(case_id=case.internal_id)
        analysis_sex: dict = self._get_analysis_sex_mip_dna(qc_metrics_file)
        samples_sex: dict[str, dict[str, str]] = {}
        for sample in case.samples:
            sample_id: str = sample.internal_id
            samples_sex[sample_id] = {
                "pedigree": sample.sex,
                "analysis": analysis_sex[sample_id],
            }
        return samples_sex

    @staticmethod
    def _get_analysis_sex_mip_dna(qc_metrics_file: Path) -> dict:
        """Return analysis sex for each sample of an analysis."""
        qc_metrics: MIPMetricsDeliverables = (
            UploadGenotypesAPI._get_parsed_qc_metrics_deliverables_mip_dna(qc_metrics_file)
        )
        return {
            sample_id_metric.sample_id: sample_id_metric.predicted_sex
            for sample_id_metric in qc_metrics.sample_id_metrics
        }

    def _get_qcmetrics_file(self, case_id: str) -> Path:
        """Return a QC metrics file path.
        Raises: FileNotFoundError if nothing is found in the Housekeeper bundle."""
        tags: set[str] = HkAnalysisMetricsTag.QC_METRICS
        hk_qcmetrics: File = self.hk.get_file_from_latest_version(bundle_name=case_id, tags=tags)
        if not hk_qcmetrics:
            raise FileNotFoundError("QC metrics file not found for the given Housekeeper version.")
        LOG.debug(f"Found QC metrics file {hk_qcmetrics.full_path}")
        return Path(hk_qcmetrics.full_path)

    @staticmethod
    def _get_parsed_qc_metrics_deliverables_mip_dna(
        qc_metrics_file: Path,
    ) -> MIPMetricsDeliverables:
        """Parse and return a QC metrics file."""
        qcmetrics_raw: dict = ReadFile.get_content_from_file(
            file_format=FileFormat.YAML, file_path=qc_metrics_file
        )
        return MIPMetricsDeliverables(**qcmetrics_raw)

    def _get_samples_sex_raredisease(self, case: Case) -> dict[str, dict[str, str]]:
        """Return sex information from StatusDB and from analysis prediction."""
        qc_metrics_file: Path = self._get_qcmetrics_file(case_id=case.internal_id)
        samples_sex: dict[str, dict[str, str]] = {}
        for sample in case.samples:
            sample_id: str = sample.internal_id
            samples_sex[sample_id] = {
                "pedigree": sample.sex,
                "analysis": self._get_analysis_sex_raredisease(
                    qc_metrics_file=qc_metrics_file, sample_id=sample_id
                ),
            }
        return samples_sex

    def _get_analysis_sex_raredisease(self, qc_metrics_file: Path, sample_id: str) -> str:
        """Return analysis sex for each sample of an analysis."""
        qc_metrics: list[MetricsBase] = self._get_parsed_qc_metrics_deliverables_raredisease(
            qc_metrics_file
        )
        for metric in qc_metrics:
            if metric.name == RAREDISEASE_PREDICTED_SEX_METRIC and metric.id == sample_id:
                return str(metric.value)

    @staticmethod
    def _get_parsed_qc_metrics_deliverables_raredisease(qc_metrics: Path) -> list[MetricsBase]:
        """Parse and return a QC metrics file."""
        qcmetrics_raw: dict = ReadFile.get_content_from_file(
            file_format=FileFormat.YAML, file_path=qc_metrics
        )
        return [MetricsBase(**metric) for metric in qcmetrics_raw["metrics"]]

    @staticmethod
    def _is_variant_file(genotype_file: File):
        return genotype_file.full_path.endswith(
            FileExtensions.VCF_GZ
        ) or genotype_file.full_path.endswith(FileExtensions.BCF)

    def _get_single_genotype_file(self, hk_genotype_files: list[File]) -> File:
        """
        Returns the single .bcf or .vcf.gz file expected to be found in the provided list. Raises an error if the amount of such files is different from 1.
        """
        filtered_files = [file for file in hk_genotype_files if self._is_variant_file(file)]
        if len(filtered_files) != 1:
            raise ValueError(
                f"Error: Expecte one genotype file, but found {len(filtered_files)} "
                f"({', '.join(map(str, filtered_files))})."
            )
        return filtered_files[0]



import logging

from housekeeper.store.models import File

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import Workflow
from cg.services.deliver_files.file_fetcher.abstract import (
    FetchDeliveryFilesService,
)
from cg.services.deliver_files.file_fetcher.error_handling import (
    handle_missing_bundle_errors,
)
from cg.services.deliver_files.file_fetcher.exc import NoDeliveryFilesError
from cg.services.deliver_files.file_fetcher.models import (
    CaseFile,
    DeliveryFiles,
    DeliveryMetaData,
    SampleFile,
)
from cg.services.deliver_files.tag_fetcher.abstract import (
    FetchDeliveryFileTagsService,
)
from cg.store.models import Case
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class AnalysisDeliveryFileFetcher(FetchDeliveryFilesService):
    """
    Fetch the analysis files for a case from Housekeeper.
    """

    def __init__(
        self, status_db: Store, hk_api: HousekeeperAPI, tags_fetcher: FetchDeliveryFileTagsService
    ):
        self.status_db = status_db
        self.hk_api = hk_api
        self.tags_fetcher = tags_fetcher

    def get_files_to_deliver(self, case_id: str) -> DeliveryFiles:
        """Return a list of analysis files to be delivered for a case."""
        LOG.debug(f"[FETCH SERVICE] Fetching analysis files for case: {case_id}")
        case: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        analysis_case_files: list[CaseFile] = self._get_analysis_case_delivery_files(case)
        analysis_sample_files: list[SampleFile] = self._get_analysis_sample_delivery_files(case)
        delivery_data = DeliveryMetaData(
            case_id=case.internal_id,
            customer_internal_id=case.customer.internal_id,
            ticket_id=case.latest_ticket,
        )

        return self._validate_delivery_has_content(
            DeliveryFiles(
                delivery_data=delivery_data,
                case_files=analysis_case_files,
                sample_files=analysis_sample_files,
            )
        )

    @staticmethod
    def _validate_delivery_has_content(delivery_files: DeliveryFiles) -> DeliveryFiles:
        """Check if the delivery files has files to deliver."""
        if delivery_files.case_files or delivery_files.sample_files:
            return delivery_files
        LOG.info(
            f"No files to deliver for case {delivery_files.delivery_data.case_id} in ticket: {delivery_files.delivery_data.ticket_id}"
        )
        raise NoDeliveryFilesError

    @handle_missing_bundle_errors
    def _get_sample_files_from_case_bundle(
        self, workflow: Workflow, sample_id: str, case_id: str
    ) -> list[SampleFile]:
        """Return a list of files from a case bundle with a sample id as tag."""
        sample_tags: list[set[str]] = self.tags_fetcher.fetch_tags(workflow).sample_tags
        sample_tags_with_sample_id: list[set[str]] = [tag | {sample_id} for tag in sample_tags]
        sample_files: list[File] = self.hk_api.get_files_from_latest_version_containing_tags(
            bundle_name=case_id, tags=sample_tags_with_sample_id
        )
        sample_name: str = self.status_db.get_sample_by_internal_id(sample_id).name
        return [
            SampleFile(
                case_id=case_id,
                sample_id=sample_id,
                sample_name=sample_name,
                file_path=sample_file.full_path,
            )
            for sample_file in sample_files
        ]

    def _get_analysis_sample_delivery_files(self, case: Case) -> list[SampleFile] | None:
        """Return a all sample files to deliver for a case."""
        sample_ids: list[str] = case.sample_ids
        delivery_files: list[SampleFile] = []
        for sample_id in sample_ids:
            sample_files: list[SampleFile] = self._get_sample_files_from_case_bundle(
                case_id=case.internal_id, sample_id=sample_id, workflow=case.data_analysis
            )
            delivery_files.extend(sample_files) if sample_files else None
        return delivery_files

    @handle_missing_bundle_errors
    def _get_analysis_case_delivery_files(self, case: Case) -> list[CaseFile]:
        """
        Return a complete list of analysis case files to be delivered and ignore analysis sample
        files.
        """
        case_tags: list[set[str]] = self.tags_fetcher.fetch_tags(case.data_analysis).case_tags
        sample_id_tags: list[str] = case.sample_ids
        case_files: list[File] = self.hk_api.get_files_from_latest_version_containing_tags(
            bundle_name=case.internal_id, tags=case_tags, excluded_tags=sample_id_tags
        )
        return [
            CaseFile(
                case_id=case.internal_id,
                case_name=case.name,
                file_path=case_file.full_path,
            )
            for case_file in case_files
        ]



from enum import Enum
from typing import Callable

from sqlalchemy import asc, desc, or_
from sqlalchemy.orm import Query

from cg.constants import Workflow
from cg.server.dto.orders.orders_request import OrderSortField, SortOrder
from cg.store.models import Customer, Order


def filter_orders_by_workflow(orders: Query, workflow: str | None, **kwargs) -> Query:
    if workflow == Workflow.BALSAMIC:
        return orders.filter(Order.workflow.startswith(workflow))
    return orders.filter(Order.workflow == workflow) if workflow else orders


def filter_orders_by_id(orders: Query, id: int | None, **kwargs) -> Query:
    return orders.filter(Order.id == id) if id else orders


def filter_orders_by_ids(orders: Query, ids: list[int] | None, **kwargs) -> Query:
    return orders.filter(Order.id.in_(ids))


def apply_pagination(orders: Query, page: int | None, page_size: int | None, **kwargs) -> Query:
    return orders.limit(page_size).offset((page - 1) * page_size) if page and page_size else orders


def filter_orders_by_ticket_id(orders: Query, ticket_id: int | None, **kwargs) -> Query:
    return orders.filter(Order.ticket_id == ticket_id) if ticket_id else orders


def filter_orders_by_search(orders: Query, search: str | None, **kwargs) -> Query:
    if not search:
        return orders

    orders = orders.join(Order.customer)
    return orders.filter(
        or_(
            Order.id == search,
            Order.ticket_id == search,
            Customer.internal_id.icontains(search),
        )
    )


def filter_orders_by_is_open(orders: Query, is_open: bool | None, **kwargs) -> Query:
    return orders.filter(Order.is_open == is_open) if is_open is not None else orders


def apply_sorting(
    orders: Query, sort_field: OrderSortField | None, sort_order: SortOrder | None, **kwargs
) -> Query:
    if sort_field:
        column = getattr(Order, sort_field)
        if sort_order == "asc":
            return orders.order_by(asc(column))
        return orders.order_by(desc(column))
    return orders


class OrderFilter(Enum):
    BY_ID: Callable = filter_orders_by_id
    BY_IDS: Callable = filter_orders_by_ids
    BY_SEARCH: Callable = filter_orders_by_search
    BY_TICKET_ID: Callable = filter_orders_by_ticket_id
    BY_WORKFLOW: Callable = filter_orders_by_workflow
    BY_OPEN: Callable = filter_orders_by_is_open
    PAGINATE: Callable = apply_pagination
    SORT: Callable = apply_sorting


def apply_order_filters(
    filters: list[OrderFilter],
    orders: Query,
    id: int = None,
    ids: list[int] = None,
    ticket_id: int = None,
    workflow: SortOrder = None,
    page: int = None,
    page_size: int = None,
    sort_field: OrderSortField = None,
    sort_order: SortOrder = None,
    search: str = None,
    is_open: bool = None,
) -> Query:
    for filter in filters:
        orders: Query = filter(
            orders=orders,
            id=id,
            ids=ids,
            workflow=workflow,
            page=page,
            page_size=page_size,
            ticket_id=ticket_id,
            sort_field=sort_field,
            sort_order=sort_order,
            search=search,
            is_open=is_open,
        )
    return orders



import logging

import click

from cg.cli.delete.case import delete_case
from cg.constants.cli_options import DRY_RUN
from cg.store.models import Case, Sample
from cg.store.store import Store

CONFIRM = "Continue?"

LOG = logging.getLogger(__name__)


def _get_samples_by_identifiers(identifiers: click.Tuple([str, str]), store: Store) -> [Sample]:
    """Get samples matched by given set of identifiers"""
    identifier_args = dict(identifiers)
    return store.get_samples_by_any_id(**identifier_args)


def _get_cases(identifiers: click.Tuple([str, str]), store: Store) -> [Case]:
    """Get cases that have samples that match identifiers if given"""
    samples_by_id = _get_samples_by_identifiers(identifiers, store)
    _cases = set()
    for sample in samples_by_id:
        for link in sample.links:
            _cases.add(link.case)

    return _cases


@click.command("cases")
@DRY_RUN
@click.option(
    "--sample-identifier",
    "identifiers",
    nargs=2,
    type=click.Tuple([str, str]),
    multiple=True,
    help="Give an identifier on sample and the value to use it with, e.g. --sample-identifier "
    "name Prov52",
)
@click.pass_context
def delete_cases(
    context: click.Context,
    dry_run: bool,
    identifiers: click.Tuple([str, str]),
):
    """Delete many cases of samples at the same time"""
    store = context.obj.status_db
    _cases = _get_cases(identifiers, store)

    if not _cases:
        LOG.error("No cases to delete!")
        raise click.Abort

    if dry_run:
        LOG.info("Cases (that will NOT be deleted due to --dry-run):")
    else:
        LOG.info("Would DELETE cases:")

    for _case in _cases:
        LOG.info(_case)

    if not (click.confirm(CONFIRM)):
        raise click.Abort

    for _case in _cases:
        context.invoke(
            delete_case,
            case_id=_case.internal_id,
            dry_run=dry_run,
        )



from cg.server.dto.orders.orders_response import Order, OrdersResponse, OrderSummary
from cg.store.models import Order as DatabaseOrder


def create_order_response(order: DatabaseOrder, summary: OrderSummary | None = None) -> Order:
    return Order(
        customer_id=order.customer.internal_id,
        ticket_id=order.ticket_id,
        order_date=str(order.order_date.date()),
        id=order.id,
        is_open=order.is_open,
        workflow=order.workflow,
        summary=summary,
    )


def create_orders_response(
    orders: list[DatabaseOrder], summaries: list[OrderSummary], total: int
) -> OrdersResponse:
    orders: list[Order] = [create_order_response(order) for order in orders]
    _add_summaries(orders=orders, summaries=summaries)
    return OrdersResponse(orders=orders, total_count=total)


def _add_summaries(orders: list[Order], summaries: list[OrderSummary]) -> list[Order]:
    order_map = {order.id: order for order in orders}
    for summary in summaries:
        order = order_map[summary.order_id]
        order.summary = summary
    return orders


def order_is_closed(case_count: int, delivered_analyses: int) -> bool:
    return delivered_analyses >= case_count



"""Utility functions for the Illumina post-processing service."""

import logging
import re
from pathlib import Path

from cg.constants.constants import FileExtensions
from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.constants.sequencing import FLOWCELL_Q30_THRESHOLD, Sequencers
from cg.services.illumina.data_transfer.models import (
    IlluminaSampleSequencingMetricsDTO,
)
from cg.utils.files import get_files_matching_pattern, is_pattern_in_file_path_name, rename_file

LOG = logging.getLogger(__name__)


def is_sample_negative_control_with_reads_in_lane(
    is_negative_control: bool, metric: IlluminaSampleSequencingMetricsDTO
) -> bool:
    return is_negative_control and metric.total_reads_in_lane != 0


def _is_sample_id_in_directory_name(directory: Path, sample_internal_id: str) -> bool:
    """Validate that directory name is formatted as Sample_<sample_id> or Sample_<sample_id>_."""
    sample_pattern: str = f"Sample_{sample_internal_id}"
    return f"{sample_pattern}_" in directory.name or sample_pattern == directory.name


def _is_file_path_compressed_fastq(file_path: Path) -> bool:
    return file_path.name.endswith(f"{FileExtensions.FASTQ}{FileExtensions.GZIP}")


def _is_lane_in_fastq_file_name(sample_fastq: Path) -> bool:
    """Validate that fastq contains lane number formatted as _L<lane_number>"""
    return bool(re.search(r"_L\d+", sample_fastq.name))


def _is_valid_sample_fastq_file(sample_fastq: Path, sample_internal_id: str) -> bool:
    """
    Validate that the sample fastq file name is formatted as expected.

    Assumptions:
    1. The sample fastq file name ends with .fastq.gz
    2. The sample fastq file name contains the lane number formatted as _L<lane_number>
    3. The sample internal id is present in the parent directory name or in the file name.
    """
    is_sample_id_in_directory: bool = _is_sample_id_in_directory_name(
        directory=sample_fastq.parent, sample_internal_id=sample_internal_id
    )
    is_sample_id_in_file_name: bool = is_pattern_in_file_path_name(
        file_path=sample_fastq, pattern=sample_internal_id
    )

    return (
        _is_file_path_compressed_fastq(sample_fastq)
        and _is_lane_in_fastq_file_name(sample_fastq)
        and (is_sample_id_in_directory or is_sample_id_in_file_name)
    )


def _get_valid_sample_fastqs(fastq_paths: list[Path], sample_internal_id: str) -> list[Path]:
    """Return a list of valid fastq files."""
    return [
        fastq
        for fastq in fastq_paths
        if _is_valid_sample_fastq_file(sample_fastq=fastq, sample_internal_id=sample_internal_id)
    ]


def get_sample_fastqs_from_flow_cell(
    demultiplexed_run_path: Path, sample_internal_id: str
) -> list[Path] | None:
    """Retrieve all fastq files for a specific sample in a demultiplex run directory."""

    # The flat output structure for runs demultiplexed with BCLConvert on hasta
    root_pattern = f"{sample_internal_id}_S*_L*_R*_*{FileExtensions.FASTQ}{FileExtensions.GZIP}"

    # The pattern for NovaSeqX flow cells demultiplexed on board of the dragen
    demux_on_sequencer_pattern = (
        f"BCLConvert/fastq/{sample_internal_id}"
        f"_S*_L*_R*_*{FileExtensions.FASTQ}{FileExtensions.GZIP}"
    )

    for pattern in [
        root_pattern,
        demux_on_sequencer_pattern,
    ]:
        sample_fastqs: list[Path] = get_files_matching_pattern(
            directory=demultiplexed_run_path, pattern=pattern
        )
        valid_sample_fastqs: list[Path] = _get_valid_sample_fastqs(
            fastq_paths=sample_fastqs, sample_internal_id=sample_internal_id
        )

        if valid_sample_fastqs:
            return valid_sample_fastqs


def _add_flow_cell_name_to_fastq_file_path(fastq_file_path: Path, flow_cell_id: str) -> Path:
    """Add the flow cell name to the fastq file path if missing."""
    if is_pattern_in_file_path_name(file_path=fastq_file_path, pattern=flow_cell_id):
        LOG.debug(f"Flow cell id {flow_cell_id} already in {fastq_file_path}. Skipping renaming.")
        return fastq_file_path
    LOG.debug(f"Adding flow cell id {flow_cell_id} to {fastq_file_path}.")
    return Path(fastq_file_path.parent, f"{flow_cell_id}_{fastq_file_path.name}")


def rename_fastq_file_if_needed(fastq_file_path: Path, flow_cell_name: str) -> Path:
    """Rename the given fastq file path if the renamed fastq file path does not exist."""
    renamed_fastq_file_path: Path = _add_flow_cell_name_to_fastq_file_path(
        fastq_file_path=fastq_file_path, flow_cell_id=flow_cell_name
    )
    if fastq_file_path != renamed_fastq_file_path:
        rename_file(file_path=fastq_file_path, renamed_file_path=renamed_fastq_file_path)
    return renamed_fastq_file_path


def _get_weighted_average(
    total_1: int, percentage_1: float, total_2: int, percentage_2: float
) -> float:
    """Calculate the weighted average of two percentages."""
    if total_1 == 0 and total_2 == 0:
        return 0
    return (total_1 * percentage_1 + total_2 * percentage_2) / (total_1 + total_2)


def _combine_metrics(
    existing_metric: IlluminaSampleSequencingMetricsDTO,
    new_metric: IlluminaSampleSequencingMetricsDTO,
) -> IlluminaSampleSequencingMetricsDTO:
    """Update an existing metric with data from a new metric."""

    combined_q30_percentage: float = _get_weighted_average(
        total_1=existing_metric.total_reads_in_lane,
        percentage_1=existing_metric.base_passing_q30_percent,
        total_2=new_metric.total_reads_in_lane,
        percentage_2=new_metric.base_passing_q30_percent,
    )
    combined_mean_quality_score: float = _get_weighted_average(
        total_1=existing_metric.total_reads_in_lane,
        percentage_1=existing_metric.base_mean_quality_score,
        total_2=new_metric.total_reads_in_lane,
        percentage_2=new_metric.base_mean_quality_score,
    )
    combined_yield_q30_percentage: float = _get_weighted_average(
        total_1=existing_metric.yield_,
        percentage_1=existing_metric.yield_q30,
        total_2=new_metric.yield_,
        percentage_2=new_metric.yield_q30,
    )
    combined_reads: int = existing_metric.total_reads_in_lane + new_metric.total_reads_in_lane
    combined_yield: int = existing_metric.yield_ + new_metric.yield_

    existing_metric.base_passing_q30_percent = combined_q30_percentage
    existing_metric.base_mean_quality_score = combined_mean_quality_score
    existing_metric.total_reads_in_lane = combined_reads
    existing_metric.yield_ = combined_yield
    existing_metric.yield_q30 = combined_yield_q30_percentage

    return existing_metric


def combine_sample_metrics_with_undetermined(
    sample_metrics: list[IlluminaSampleSequencingMetricsDTO],
    undetermined_metrics: list[IlluminaSampleSequencingMetricsDTO],
) -> list[IlluminaSampleSequencingMetricsDTO]:
    """Combine metrics for sample metrics with metrics for undetermined reads.
    This occurs only if a sample is the only sample in a lane.
    """

    metrics: dict[tuple[str, int], IlluminaSampleSequencingMetricsDTO] = {
        (metric.sample_id, metric.flow_cell_lane): metric for metric in sample_metrics
    }
    for undetermined_metric in undetermined_metrics:
        key = (
            undetermined_metric.sample_id,
            undetermined_metric.flow_cell_lane,
        )
        existing_metric: IlluminaSampleSequencingMetricsDTO = metrics.get(key)

        if existing_metric:
            combined_metric: IlluminaSampleSequencingMetricsDTO = _combine_metrics(
                existing_metric=existing_metric, new_metric=undetermined_metric
            )
            metrics[key] = combined_metric
        else:
            metrics[key] = undetermined_metric
    return list(metrics.values())


def get_undetermined_fastqs(lane: int, demultiplexed_run_path: Path) -> list[Path]:
    """Get the undetermined fastq files for a specific lane on a flow cell."""
    undetermined_pattern = f"Undetermined*_L00{lane}_*{FileExtensions.FASTQ}{FileExtensions.GZIP}"
    undetermined_in_root: list[Path] = get_files_matching_pattern(
        directory=demultiplexed_run_path,
        pattern=undetermined_pattern,
    )
    return undetermined_in_root


def create_delivery_file_in_flow_cell_directory(flow_cell_directory: Path) -> None:
    Path(flow_cell_directory, DemultiplexingDirsAndFiles.DELIVERY).touch()


def get_lane_from_sample_fastq(sample_fastq_path: Path) -> int:
    """
    Extract the lane number from the sample fastq path.
    Pre-condition:
        - The fastq file name contains the lane number formatted as _L<lane_number>
    """
    pattern = r"_L(\d+)"
    lane_match = re.search(pattern, sample_fastq_path.name)

    if lane_match:
        return int(lane_match.group(1))

    raise ValueError(f"Could not extract lane number from fastq file name {sample_fastq_path.name}")


def get_q30_threshold(sequencer_type: Sequencers) -> int:
    return FLOWCELL_Q30_THRESHOLD[sequencer_type]



from cg.constants.constants import MutantQC
from cg.services.sequencing_qc_service.quality_checks.utils import sample_has_enough_reads
from cg.store.models import Sample


def has_sample_valid_total_reads(
    sample: Sample,
) -> bool:
    return sample_has_enough_reads(sample=sample)


def has_internal_negative_control_sample_valid_total_reads(
    sample: Sample,
) -> bool:
    return sample.reads < MutantQC.INTERNAL_NEGATIVE_CONTROL_READS_THRESHOLD


def has_external_negative_control_sample_valid_total_reads(
    sample: Sample,
) -> bool:
    return sample.reads < MutantQC.EXTERNAL_NEGATIVE_CONTROL_READS_THRESHOLD



"""Module for Raredisease Analysis API."""

import logging
from itertools import permutations
from pathlib import Path
from typing import Any

from housekeeper.store.models import File

from cg.clients.chanjo2.models import (
    CoverageMetrics,
    CoveragePostRequest,
    CoveragePostResponse,
    CoverageSample,
)
from cg.constants import DEFAULT_CAPTURE_KIT, Workflow
from cg.constants.constants import AnalysisType, GenomeVersion
from cg.constants.nf_analysis import (
    RAREDISEASE_COVERAGE_FILE_TAGS,
    RAREDISEASE_COVERAGE_INTERVAL_TYPE,
    RAREDISEASE_COVERAGE_THRESHOLD,
    RAREDISEASE_METRIC_CONDITIONS,
    RAREDISEASE_PARENT_PEDDY_METRIC_CONDITION,
)
from cg.constants.scout import RAREDISEASE_CASE_TAGS, ScoutExportFileName
from cg.constants.subject import PlinkPhenotypeStatus, PlinkSex
from cg.meta.workflow.nf_analysis import NfAnalysisAPI
from cg.models.analysis import NextflowAnalysis
from cg.models.cg_config import CGConfig
from cg.models.deliverables.metric_deliverables import MetricsBase, MultiqcDataJson
from cg.models.raredisease.raredisease import (
    RarediseaseParameters,
    RarediseaseSampleSheetEntry,
    RarediseaseSampleSheetHeaders,
)
from cg.resources import RAREDISEASE_BUNDLE_FILENAMES_PATH
from cg.store.models import CaseSample, Sample

LOG = logging.getLogger(__name__)


class RarediseaseAnalysisAPI(NfAnalysisAPI):
    """Handles communication between RAREDISEASE processes
    and the rest of CG infrastructure."""

    def __init__(
        self,
        config: CGConfig,
        workflow: Workflow = Workflow.RAREDISEASE,
    ):
        super().__init__(config=config, workflow=workflow)
        self.root_dir: str = config.raredisease.root
        self.nfcore_workflow_path: str = config.raredisease.workflow_path
        self.references: str = config.raredisease.references
        self.profile: str = config.raredisease.profile
        self.conda_env: str = config.raredisease.conda_env
        self.conda_binary: str = config.raredisease.conda_binary
        self.config_platform: str = config.raredisease.config_platform
        self.config_params: str = config.raredisease.config_params
        self.config_resources: str = config.raredisease.config_resources
        self.tower_binary_path: str = config.tower_binary_path
        self.tower_workflow: str = config.raredisease.tower_workflow
        self.account: str = config.raredisease.slurm.account
        self.email: str = config.raredisease.slurm.mail_user
        self.compute_env_base: str = config.raredisease.compute_env
        self.revision: str = config.raredisease.revision
        self.nextflow_binary_path: str = config.raredisease.binary_path

    @property
    def sample_sheet_headers(self) -> list[str]:
        """Headers for sample sheet."""
        return RarediseaseSampleSheetHeaders.list()

    def get_sample_sheet_content_per_sample(self, case_sample: CaseSample) -> list[list[str]]:
        """Collect and format information required to build a sample sheet for a single sample."""
        fastq_forward_read_paths, fastq_reverse_read_paths = self.get_paired_read_paths(
            sample=case_sample.sample
        )
        sample_sheet_entry = RarediseaseSampleSheetEntry(
            name=case_sample.sample.internal_id,
            fastq_forward_read_paths=fastq_forward_read_paths,
            fastq_reverse_read_paths=fastq_reverse_read_paths,
            sex=self.get_sex_code(case_sample.sample.sex),
            phenotype=self.get_phenotype_code(case_sample.status),
            paternal_id=case_sample.get_paternal_sample_id,
            maternal_id=case_sample.get_maternal_sample_id,
            case_id=case_sample.case.internal_id,
        )
        return sample_sheet_entry.reformat_sample_content

    @property
    def is_gene_panel_required(self) -> bool:
        """Return True if a gene panel is needs to be created using the information in StatusDB and exporting it from Scout."""
        return True

    def get_target_bed(self, case_id: str, analysis_type: str) -> str:
        """
        Return the target bed file from LIMS and use default capture kit for WGS.
        """
        target_bed: str = self.get_target_bed_from_lims(case_id=case_id)
        if not target_bed:
            if analysis_type == AnalysisType.WHOLE_GENOME_SEQUENCING:
                return DEFAULT_CAPTURE_KIT
            raise ValueError("No capture kit was found in LIMS")
        return target_bed

    def get_germlinecnvcaller_flag(self, analysis_type: str) -> bool:
        if analysis_type == AnalysisType.WHOLE_GENOME_SEQUENCING:
            return True
        return False

    def get_workflow_parameters(self, case_id: str) -> RarediseaseParameters:
        """Return parameters."""
        analysis_type: AnalysisType = self.get_data_analysis_type(case_id=case_id)
        target_bed: str = self.get_target_bed(case_id=case_id, analysis_type=analysis_type)
        skip_germlinecnvcaller = self.get_germlinecnvcaller_flag(analysis_type=analysis_type)
        outdir = self.get_case_path(case_id=case_id)

        return RarediseaseParameters(
            local_genomes=str(self.references),
            input=self.get_sample_sheet_path(case_id=case_id),
            outdir=outdir,
            analysis_type=analysis_type,
            target_bed=Path(self.references, target_bed).as_posix(),
            save_mapped_as_cram=True,
            skip_germlinecnvcaller=skip_germlinecnvcaller,
            vcfanno_extra_resources=f"{outdir}/{ScoutExportFileName.MANAGED_VARIANTS}",
            vep_filters_scout_fmt=f"{outdir}/{ScoutExportFileName.PANELS}",
        )

    @staticmethod
    def get_phenotype_code(phenotype: str) -> int:
        """Return Raredisease phenotype code."""
        LOG.debug("Translate phenotype to integer code")
        try:
            code = PlinkPhenotypeStatus[phenotype.upper()]
        except KeyError:
            raise ValueError(f"{phenotype} is not a valid phenotype")
        return code

    @staticmethod
    def get_sex_code(sex: str) -> int:
        """Return Raredisease sex code."""
        LOG.debug("Translate sex to integer code")
        try:
            code = PlinkSex[sex.upper()]
        except KeyError:
            raise ValueError(f"{sex} is not a valid sex")
        return code

    @staticmethod
    def get_bundle_filenames_path() -> Path:
        """Return Raredisease bundle filenames path."""
        return RAREDISEASE_BUNDLE_FILENAMES_PATH

    @property
    def is_managed_variants_required(self) -> bool:
        """Return True if a managed variants needs to be exported from Scout."""
        return True

    @property
    def root(self) -> str:
        return self.config.raredisease.root

    def write_managed_variants(self, case_id: str, content: list[str]) -> None:
        self._write_managed_variants(out_dir=Path(self.root, case_id), content=content)

    def get_managed_variants(self, case_id: str) -> list[str]:
        """Create and return the managed variants."""
        return self._get_managed_variants(
            genome_build=self.get_gene_panel_genome_build(case_id=case_id)
        )

    def get_workflow_metrics(self, sample_id: str) -> dict:
        """Return Raredisease workflow metric conditions for a sample."""
        sample: Sample = self.status_db.get_sample_by_internal_id(internal_id=sample_id)
        if "-" not in sample_id:
            metric_conditions: dict[str, dict[str, Any]] = RAREDISEASE_METRIC_CONDITIONS.copy()
            self.set_order_sex_for_sample(sample, metric_conditions)
        else:
            metric_conditions = RAREDISEASE_PARENT_PEDDY_METRIC_CONDITION.copy()
        return metric_conditions

    def _get_sample_pair_patterns(self, case_id: str) -> list[str]:
        """Return sample-pair patterns for searching in MultiQC."""
        sample_ids: list[str] = list(self.status_db.get_sample_ids_by_case_id(case_id=case_id))
        pairwise_patterns: list[str] = [
            f"{sample1}-{sample2}" for sample1, sample2 in permutations(sample_ids, 2)
        ]
        return pairwise_patterns

    def get_parent_error_ped_check_metric(
        self, pair_sample_ids: str, multiqc_raw_data: dict[dict]
    ) -> MetricsBase | None:
        """Return the parsed metrics for pedigree error given a concatenated pair of sample ids."""
        metric_name: str = "parent_error_ped_check"
        peddy_metrics: dict[str, dict] = multiqc_raw_data["multiqc_peddy"]
        if sample_pair_metrics := peddy_metrics.get(pair_sample_ids, None):
            return self.get_multiqc_metric(
                metric_name=metric_name,
                metric_value=sample_pair_metrics[metric_name],
                metric_id=pair_sample_ids,
            )

    def get_raredisease_multiqc_json_metrics(self, case_id: str) -> list[MetricsBase]:
        """Return a list of the metrics specified in a MultiQC json file."""
        multiqc_json: MultiqcDataJson = self.get_multiqc_data_json(case_id=case_id)
        metrics = []
        for search_pattern, metric_id in self.get_multiqc_search_patterns(case_id).items():
            metrics_for_pattern: list[MetricsBase] = (
                self.get_metrics_from_multiqc_json_with_pattern(
                    search_pattern=search_pattern,
                    multiqc_json=multiqc_json,
                    metric_id=metric_id,
                    exact_match=self.is_multiqc_pattern_search_exact,
                )
            )
            metrics.extend(metrics_for_pattern)
        for sample_pair in self._get_sample_pair_patterns(case_id):
            if parent_error_metric := self.get_parent_error_ped_check_metric(
                pair_sample_ids=sample_pair, multiqc_raw_data=multiqc_json.report_saved_raw_data
            ):
                metrics.append(parent_error_metric)
        metrics = self.get_deduplicated_metrics(metrics=metrics)
        return metrics

    def create_metrics_deliverables_content(self, case_id: str) -> dict[str, list[dict[str, Any]]]:
        """Create the content of a Raredisease metrics deliverables file."""
        metrics: list[MetricsBase] = self.get_raredisease_multiqc_json_metrics(case_id=case_id)
        self.ensure_mandatory_metrics_present(metrics=metrics)
        return {"metrics": [metric.dict() for metric in metrics]}

    @staticmethod
    def set_order_sex_for_sample(sample: Sample, metric_conditions: dict) -> None:
        metric_conditions["predicted_sex_sex_check"]["threshold"] = sample.sex
        metric_conditions["gender"]["threshold"] = sample.sex

    def get_sample_coverage_file_path(self, bundle_name: str, sample_id: str) -> str | None:
        """Return the Raredisease d4 coverage file path."""
        coverage_file_tags: list[str] = RAREDISEASE_COVERAGE_FILE_TAGS + [sample_id]
        coverage_file: File | None = self.housekeeper_api.get_file_from_latest_version(
            bundle_name=bundle_name, tags=coverage_file_tags
        )
        if coverage_file:
            return coverage_file.full_path
        LOG.warning(f"No coverage file found with the tags: {coverage_file_tags}")
        return None

    def get_sample_coverage(
        self, case_id: str, sample_id: str, gene_ids: list[int]
    ) -> CoverageMetrics | None:
        """Return sample coverage metrics from Chanjo2."""
        genome_version: GenomeVersion = self.get_genome_build(case_id)
        coverage_file_path: str | None = self.get_sample_coverage_file_path(
            bundle_name=case_id, sample_id=sample_id
        )
        try:
            post_request = CoveragePostRequest(
                build=self.translate_genome_reference(genome_version),
                coverage_threshold=RAREDISEASE_COVERAGE_THRESHOLD,
                hgnc_gene_ids=gene_ids,
                interval_type=RAREDISEASE_COVERAGE_INTERVAL_TYPE,
                samples=[CoverageSample(coverage_file_path=coverage_file_path, name=sample_id)],
            )
            post_response: CoveragePostResponse = self.chanjo2_api.get_coverage(post_request)
            return post_response.get_sample_coverage_metrics(sample_id)
        except Exception as error:
            LOG.error(f"Error getting coverage for sample '{sample_id}', error: {error}")
            return None

    def get_scout_upload_case_tags(self) -> dict:
        """Return Raredisease Scout upload case tags."""
        return RAREDISEASE_CASE_TAGS

    def get_genome_build(self, case_id: str) -> GenomeVersion:
        """Return reference genome for a raredisease case. Currently fixed for hg19."""
        return GenomeVersion.HG19



import logging
from functools import wraps

from cg.exc import CgError
from cg.services.deliver_files.deliver_files_service.exc import DeliveryTypeNotSupported

LOG = logging.getLogger(__name__)


def handle_delivery_type_errors(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except DeliveryTypeNotSupported:
            LOG.info("Delivery type not supported or no delivery needed. Skipping delivery.")
            return

    return wrapper



from pathlib import Path

from cg.services.run_devices.run_names.pacbio import PacbioRunNamesService


def test_pacbio_get_run_names(
    pacbio_run_names_service: PacbioRunNamesService, pacbio_run_names: set
):
    """Test that getting PacBio run names works."""
    # GIVEN a PacBio run names service

    # WHEN getting the run names
    run_names: list[str] = pacbio_run_names_service.get_run_names()

    # THEN the run names are returned
    assert set(run_names) == pacbio_run_names


def test_pacbio_get_run_names_empty(tmp_path: Path):
    # GIVEN a run directory with no run folders and a PacBio run names service
    pacbio_run_names_service: PacbioRunNamesService = PacbioRunNamesService(
        run_directory=tmp_path.as_posix()
    )

    # WHEN getting the run names
    run_names: list[str] = pacbio_run_names_service.get_run_names()

    # THEN no run names are returned
    assert run_names == []


def test_pacbio_get_run_names_with_files_return_emty(tmp_path: Path):
    # GIVEN a run directory with files and a PacBio run names service
    run_dir = Path(tmp_path, "run")
    run_dir.mkdir()
    Path(run_dir, "file1").touch()
    Path(run_dir, "file2").touch()
    pacbio_run_names_service: PacbioRunNamesService = PacbioRunNamesService(
        run_directory=run_dir.as_posix()
    )

    # WHEN getting the run names
    run_names: list[str] = pacbio_run_names_service.get_run_names()

    # THEN no run names are returned
    assert run_names == []


def test_pacbio_get_run_names_with_subfolders_return_empty(tmp_path: Path):
    # GIVEN a run directory with an empty subfolder and a PacBio run names service
    run_dir = Path(tmp_path, "run")
    run_dir.mkdir()
    subfolder = Path(run_dir, "subfolder")
    subfolder.mkdir()
    pacbio_run_names_service: PacbioRunNamesService = PacbioRunNamesService(
        run_directory=run_dir.as_posix()
    )

    # WHEN getting the run names
    run_names: list[str] = pacbio_run_names_service.get_run_names()

    # THEN no run names are returned
    assert run_names == []



import os
from pathlib import Path

from cg.services.deliver_files.file_fetcher.models import CaseFile
from cg.services.deliver_files.file_formatter.models import FormattedFile


class CaseFileFormatter:

    def format_files(
        self, moved_files: list[CaseFile], ticket_dir_path: Path
    ) -> list[FormattedFile]:
        """Format the case files to deliver and return the formatted files."""
        self._create_case_name_folder(
            ticket_path=ticket_dir_path, case_name=moved_files[0].case_name
        )
        return self._format_case_files(moved_files)

    def _format_case_files(self, case_files: list[CaseFile]) -> list[FormattedFile]:
        formatted_files: list[FormattedFile] = self._get_formatted_files(case_files)
        for formatted_file in formatted_files:
            os.rename(src=formatted_file.original_path, dst=formatted_file.formatted_path)
        return formatted_files

    @staticmethod
    def _create_case_name_folder(ticket_path: Path, case_name: str) -> None:
        case_dir_path = Path(ticket_path, case_name)
        case_dir_path.mkdir(exist_ok=True)

    @staticmethod
    def _get_formatted_files(case_files: list[CaseFile]) -> list[FormattedFile]:
        """
        Returns formatted files:
        1. Adds a folder with case name to the path of the case files.
        2. Replaces case id by case name.
        """
        formatted_files: list[FormattedFile] = []
        for case_file in case_files:
            replaced_case_file_name: str = case_file.file_path.name.replace(
                case_file.case_id, case_file.case_name
            )
            formatted_file_path = Path(
                case_file.file_path.parent, case_file.case_name, replaced_case_file_name
            )
            formatted_files.append(
                FormattedFile(original_path=case_file.file_path, formatted_path=formatted_file_path)
            )
        return formatted_files



"""File includes api to uploading data into Scout."""

import logging
from pathlib import Path

from housekeeper.store.models import File, Version
from pydantic.dataclasses import dataclass

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.apps.madeline.api import MadelineAPI
from cg.apps.scout.scoutapi import ScoutAPI
from cg.constants import HK_MULTIQC_HTML_TAG, Workflow
from cg.constants.constants import FileFormat, GenomeVersion, PrepCategory
from cg.constants.housekeeper_tags import (
    HK_DELIVERY_REPORT_TAG,
    AlignmentFileTag,
    AnalysisTag,
)
from cg.constants.scout import ScoutCustomCaseReportTags
from cg.exc import CgDataError, HousekeeperBundleVersionMissingError
from cg.io.controller import WriteFile
from cg.meta.upload.scout.balsamic_config_builder import BalsamicConfigBuilder
from cg.meta.upload.scout.balsamic_umi_config_builder import BalsamicUmiConfigBuilder
from cg.meta.upload.scout.mip_config_builder import MipConfigBuilder
from cg.meta.upload.scout.raredisease_config_builder import RarediseaseConfigBuilder
from cg.meta.upload.scout.rnafusion_config_builder import RnafusionConfigBuilder
from cg.meta.upload.scout.scout_config_builder import ScoutConfigBuilder
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.utils.genome_build_helpers import (
    genome_to_scout_format,
    get_genome_build,
)
from cg.models.scout.scout_load_config import ScoutLoadConfig
from cg.store.models import Analysis, Case, Customer, Sample
from cg.store.store import Store

LOG = logging.getLogger(__name__)


@dataclass
class RNADNACollection:
    """Contains the id for an RNA sample, the name of its connected DNA sample,
    and a list of connected, uploaded DNA cases."""

    rna_sample_internal_id: str
    dna_sample_name: str
    dna_case_ids: list[str]


class UploadScoutAPI:
    """Class that handles everything that has to do with uploading to Scout."""

    def __init__(
        self,
        hk_api: HousekeeperAPI,
        scout_api: ScoutAPI,
        lims_api: LimsAPI,
        analysis_api: AnalysisAPI,
        madeline_api: MadelineAPI,
        status_db: Store,
    ):
        self.housekeeper = hk_api
        self.scout_api = scout_api
        self.madeline_api = madeline_api
        self.analysis_api = analysis_api
        self.raredisease_analysis_api = analysis_api
        self.lims = lims_api
        self.status_db = status_db

    def generate_config(self, analysis: Analysis) -> ScoutLoadConfig:
        """Fetch data about an analysis to load Scout."""
        LOG.info("Generate scout load config")
        # Fetch last version from housekeeper
        # This should be safe since analyses are only added if data is analysed
        hk_version_obj: Version = self.housekeeper.last_version(analysis.case.internal_id)
        LOG.debug(f"Found housekeeper version {hk_version_obj.id}")

        LOG.info(f"Found workflow {analysis.workflow}")
        config_builder = self.get_config_builder(analysis=analysis, hk_version=hk_version_obj)
        return config_builder.build_load_config()

    @staticmethod
    def get_load_config_tag() -> str:
        """Get the Housekeeper tag for a Scout load config."""
        return "scout-load-config"

    @staticmethod
    def save_config_file(upload_config: ScoutLoadConfig, file_path: Path) -> None:
        """Save a Scout load config file to the supplied file path."""

        LOG.info(f"Save Scout load config to {file_path.as_posix()}.")
        WriteFile.write_file_from_content(
            content=upload_config.model_dump(exclude_none=True),
            file_format=FileFormat.YAML,
            file_path=file_path,
        )

    def add_scout_config_to_hk(
        self, config_file_path: Path, case_id: str, delete: bool = False
    ) -> File:
        """Add Scout load config to Housekeeper bundle."""
        LOG.info(f"Adding load config {config_file_path} to Housekeeper.")
        tag_name: str = self.get_load_config_tag()
        version: Version = self.housekeeper.last_version(case_id)
        uploaded_config_file: File | None = self.housekeeper.get_latest_file_from_version(
            version=version, tags={tag_name}
        )
        if uploaded_config_file:
            LOG.info(f"Found config file: {uploaded_config_file}.")
            if not delete:
                raise FileExistsError("Upload config already exists.")
            self.housekeeper.delete_file(uploaded_config_file.id)

        file_obj: File = self.housekeeper.add_file(
            path=str(config_file_path), version_obj=version, tags=[tag_name]
        )
        self.housekeeper.include_file(file_obj=file_obj, version_obj=version)
        self.housekeeper.add_commit(file_obj)

        LOG.info(f"Added Scout load config to Housekeeper: {config_file_path}.")
        return file_obj

    def get_multiqc_html_report(
        self, case_id: str, workflow: Workflow
    ) -> tuple[ScoutCustomCaseReportTags, File | None]:
        """Return a multiqc report for a case in Housekeeper."""
        if workflow == Workflow.MIP_RNA:
            tags: set[str] = {HK_MULTIQC_HTML_TAG, case_id}
            return (
                ScoutCustomCaseReportTags.MULTIQC_RNA,
                self.housekeeper.get_file_from_latest_version(bundle_name=case_id, tags=tags),
            )
        tags: set[str] = {HK_MULTIQC_HTML_TAG, case_id}
        if workflow == Workflow.TOMTE:
            return (
                ScoutCustomCaseReportTags.MULTIQC_RNA,
                self.housekeeper.get_file_from_latest_version(bundle_name=case_id, tags=tags),
            )
        return ScoutCustomCaseReportTags.MULTIQC, self.housekeeper.get_file_from_latest_version(
            bundle_name=case_id, tags=tags
        )

    def get_fusion_report(self, case_id: str, research: bool) -> File | None:
        """Return a fusion report for a case in housekeeper."""

        tags = {AnalysisTag.FUSION}
        if research:
            tags.add(AnalysisTag.RESEARCH)
        else:
            tags.add(AnalysisTag.CLINICAL)

        return self.housekeeper.get_file_from_latest_version(bundle_name=case_id, tags=tags)

    def get_rna_delivery_report(self, case_id: str) -> File | None:
        """Return a RNA report for a case in housekeeper."""

        tags = {HK_DELIVERY_REPORT_TAG}

        return self.housekeeper.get_file_from_latest_version(bundle_name=case_id, tags=tags)

    def get_splice_junctions_bed(self, case_id: str, sample_id: str) -> File | None:
        """Return a splice junctions bed file for a case in Housekeeper."""

        tags: set[str] = {AnalysisTag.JUNCTION, AnalysisTag.BED, sample_id}
        splice_junctions_bed: File | None = None
        try:
            splice_junctions_bed = self.housekeeper.get_file_from_latest_version(
                bundle_name=case_id, tags=tags
            )
        except HousekeeperBundleVersionMissingError:
            LOG.debug("Could not find bundle for case %s", case_id)

        return splice_junctions_bed

    def get_rna_coverage_bigwig(self, case_id: str, sample_id: str) -> File | None:
        """Return an RNA coverage bigwig file for a case in Housekeeper."""
        tags: set[str] = {AnalysisTag.COVERAGE, AnalysisTag.BIGWIG, sample_id}
        return self.housekeeper.get_file_from_latest_version(bundle_name=case_id, tags=tags)

    def get_rna_omics_fraser(self, case_id: str) -> File | None:
        """Return an fraser file for a case in Housekeeper."""
        tags: set[str] = {AnalysisTag.FRASER, case_id, AnalysisTag.CLINICAL}
        return self.housekeeper.get_file_from_latest_version(bundle_name=case_id, tags=tags)

    def get_rna_omics_outrider(self, case_id: str) -> File | None:
        """Return an outrider file for a case in Housekeeper."""
        tags: set[str] = {AnalysisTag.OUTRIDER, case_id, AnalysisTag.CLINICAL}
        return self.housekeeper.get_file_from_latest_version(bundle_name=case_id, tags=tags)

    def get_unique_dna_cases_related_to_rna_case(self, case_id: str) -> set[str]:
        """Return a set of unique DNA cases related to an RNA case."""
        case: Case = self.status_db.get_case_by_internal_id(case_id)
        rna_dna_collections: list[RNADNACollection] = self.create_rna_dna_collections(case)
        unique_dna_cases_related_to_rna_case: set[str] = set()
        for rna_dna_collection in rna_dna_collections:
            unique_dna_cases_related_to_rna_case.update(rna_dna_collection.dna_case_ids)
        return unique_dna_cases_related_to_rna_case

    def get_rna_alignment_cram(self, case_id: str, sample_id: str) -> File | None:
        """Return an RNA alignment CRAM file for a case in Housekeeper."""
        tags: set[str] = {AlignmentFileTag.CRAM, sample_id}
        rna_alignment_cram: File | None = None
        try:
            rna_alignment_cram = self.housekeeper.get_file_from_latest_version(
                bundle_name=case_id, tags=tags
            )
        except HousekeeperBundleVersionMissingError:
            LOG.warning(f"Could not find bundle for {case_id}")
        return rna_alignment_cram

    def upload_rna_alignment_file(self, case_id: str, dry_run: bool) -> None:
        """Upload RNA alignment file to Scout."""
        rna_case: Case = self.status_db.get_case_by_internal_id(case_id)
        rna_dna_collections: list[RNADNACollection] = self.create_rna_dna_collections(rna_case)
        for rna_dna_collection in rna_dna_collections:
            rna_sample_internal_id: str = rna_dna_collection.rna_sample_internal_id
            dna_sample_name: str = rna_dna_collection.dna_sample_name
            rna_alignment_cram: File | None = self.get_rna_alignment_cram(
                case_id=case_id, sample_id=rna_sample_internal_id
            )
            if not rna_alignment_cram:
                raise FileNotFoundError(
                    f"No RNA alignment CRAM file was found in Housekeeper for {rna_sample_internal_id}"
                )
            LOG.debug(f"RNA alignment CRAM file {rna_alignment_cram.path} found")
            for dna_case_id in rna_dna_collection.dna_case_ids:
                LOG.info(
                    f"Uploading RNA alignment CRAM file for sample {dna_sample_name} to case {dna_case_id}"
                )
                if dry_run:
                    continue
                self.scout_api.upload_rna_alignment_file(
                    case_id=dna_case_id,
                    customer_sample_id=dna_sample_name,
                    file_path=rna_alignment_cram.full_path,
                )
        for upload_statement in self.get_rna_alignment_file_upload_summary(rna_dna_collections):
            LOG.info(upload_statement)
        LOG.info("Upload RNA alignment CRAM file finished!")

    def upload_fusion_report_to_scout(
        self, dry_run: bool, case_id: str, research: bool = False
    ) -> None:
        """Upload fusion report file for a case to Scout."""

        report_type: str = "Research" if research else "Clinical"

        fusion_report: File | None = self.get_fusion_report(case_id, research)
        if not fusion_report:
            raise FileNotFoundError(
                f"{report_type} fusion report was not found in Housekeeper for {case_id}."
            )

        LOG.info(f"{report_type} fusion report {fusion_report.path} found")

        related_dna_cases: set[str] = self.get_related_uploaded_dna_cases(case_id)
        if not related_dna_cases:
            raise CgDataError("No connected DNA case found.")

        for dna_case_id in related_dna_cases:
            LOG.info(f"Uploading {report_type} fusion report to Scout for case {dna_case_id}.")

            if dry_run:
                continue
            self.scout_api.upload_fusion_report(
                case_id=dna_case_id,
                report_path=fusion_report.full_path,
                research=research,
            )
            LOG.info(f"Uploaded {report_type} fusion report.")

        LOG.info(f"Upload {report_type} fusion report finished!")

    def upload_rna_delivery_report_to_scout(self, dry_run: bool, case_id: str) -> None:
        """Upload rna delivery report file for a case to Scout."""

        rna_delivery_report: File | None = self.get_rna_delivery_report(case_id)
        if not rna_delivery_report:
            raise FileNotFoundError(
                f"Rna delivery report was not found in Housekeeper for {case_id}."
            )

        LOG.info(f"Rna delivery report {rna_delivery_report.path} found")

        related_dna_cases: set[str] = self.get_related_uploaded_dna_cases(case_id)
        if not related_dna_cases:
            raise CgDataError("No connected DNA case has been uploaded.")

        for dna_case_id in related_dna_cases:
            LOG.debug(f"Uploading rna delivery report to Scout for case {dna_case_id}.")

            if dry_run:
                LOG.debug(
                    f"Dry run - Would have uploaded rna delivery report to Scout for case {dna_case_id}."
                )
                continue
            self.scout_api.upload_rna_delivery_report(
                case_id=dna_case_id,
                report_path=rna_delivery_report.full_path,
            )
            LOG.debug(f"Uploaded rna delivery report for {dna_case_id}.")

        LOG.info(f"Upload rna delivery report finished for case {case_id}!")

    def upload_rna_report_to_dna_case_in_scout(
        self,
        dry_run: bool,
        report_type: str,
        report_file: File,
        rna_case_id: str,
    ) -> None:
        """Upload report file to DNA cases related to an RNA case in Scout."""
        LOG.info(f"Finding DNA cases related to RNA case {rna_case_id}")

        related_dna_cases: set[str] = self.get_related_uploaded_dna_cases(rna_case_id)
        if not related_dna_cases:
            raise CgDataError("No connected DNA case has been uploaded.")
        for dna_case_id in related_dna_cases:
            LOG.info(f"Running upload of report to DNA case {dna_case_id}.")
            self.upload_report_to_scout(
                dry_run=dry_run,
                report_type=report_type,
                report_file=report_file,
                case_id=dna_case_id,
            )

    def upload_report_to_scout(
        self,
        dry_run: bool,
        case_id: str,
        report_type: str,
        report_file: File,
    ) -> None:
        """Upload report file for a case to Scout."""

        LOG.info(f"Uploading {report_type} report to Scout for case {case_id}.")

        if dry_run:
            LOG.info(f"Would have uploaded {report_type} report.")
            return
        self.scout_api.upload_report(
            case_id=case_id,
            report_path=report_file.full_path,
            report_type=report_type,
        )
        LOG.info(f"Uploaded {report_type} report.")
        LOG.info(f"Upload {report_type} report finished!")

    def upload_rna_coverage_bigwig_to_scout(self, case_id: str, dry_run: bool) -> None:
        """Upload rna_coverage_bigwig file for a case to Scout."""

        status_db: Store = self.status_db
        rna_case = status_db.get_case_by_internal_id(case_id)
        rna_dna_collections: list[RNADNACollection] = self.create_rna_dna_collections(rna_case)
        for rna_dna_collection in rna_dna_collections:
            rna_sample_internal_id: str = rna_dna_collection.rna_sample_internal_id
            dna_sample_name: str = rna_dna_collection.dna_sample_name
            rna_coverage_bigwig: File | None = self.get_rna_coverage_bigwig(
                case_id=case_id, sample_id=rna_sample_internal_id
            )

            if not rna_coverage_bigwig:
                raise FileNotFoundError(
                    f"No RNA coverage bigwig file was found in housekeeper for {rna_sample_internal_id}."
                )

            LOG.debug(f"RNA coverage bigwig file {rna_coverage_bigwig.path} found.")
            for dna_case_id in rna_dna_collection.dna_case_ids:
                LOG.info(
                    f"Uploading RNA coverage bigwig file for sample {dna_sample_name} "
                    f"in case {dna_case_id} in Scout."
                )

                if dry_run:
                    continue

                self.scout_api.upload_rna_coverage_bigwig(
                    file_path=rna_coverage_bigwig.full_path,
                    case_id=dna_case_id,
                    customer_sample_id=dna_sample_name,
                )
        for upload_statement in self.get_rna_bigwig_coverage_upload_summary(rna_dna_collections):
            LOG.info(upload_statement)
        LOG.info("Upload RNA coverage bigwig file finished!")

    def upload_omics_sample_id_to_scout(
        self, dry_run: bool, rna_dna_collections: list[RNADNACollection]
    ) -> None:
        for rna_dna_collection in rna_dna_collections:
            rna_sample_internal_id: str = rna_dna_collection.rna_sample_internal_id
            dna_sample_name: str = rna_dna_collection.dna_sample_name
            for dna_case_id in rna_dna_collection.dna_case_ids:
                LOG.info(
                    f"Uploading omics sample id for sample {dna_sample_name} "
                    f"in case {dna_case_id} in Scout."
                )
                if dry_run:
                    continue
                self.scout_api.upload_omics_sample_id(
                    dna_case_id=dna_case_id,
                    customer_sample_id=dna_sample_name,
                    rna_sample_internal_id=rna_sample_internal_id,
                )

    def upload_rna_fraser_outrider_to_scout(
        self, dry_run: bool, case_id: str, rna_dna_collections: list[RNADNACollection], cust_id: str
    ) -> None:
        """Upload omics fraser and outrider file for a case to Scout."""
        status_db: Store = self.status_db
        for rna_dna_collection in rna_dna_collections:
            rna_sample_internal_id: str = rna_dna_collection.rna_sample_internal_id
            dna_sample_name: str = rna_dna_collection.dna_sample_name
            rna_fraser: File | None = self.get_rna_omics_fraser(case_id=case_id)
            rna_outrider: File | None = self.get_rna_omics_outrider(case_id=case_id)
            if not rna_fraser:
                raise FileNotFoundError(
                    f"No RNA fraser file was found in housekeeper for {rna_sample_internal_id}."
                )
            if not rna_outrider:
                raise FileNotFoundError(
                    f"No RNA outrider file was found in housekeeper for {rna_sample_internal_id}."
                )
            LOG.debug(f"RNA fraser file {rna_fraser.path} found.")
            LOG.debug(f"RNA outrider file {rna_outrider.path} found.")
            for dna_case_id in rna_dna_collection.dna_case_ids:
                LOG.info(
                    f"Uploading RNA fraser and outrider files for sample {dna_sample_name} "
                    f"in case {dna_case_id} in Scout."
                )
                customer_case = status_db.get_case_by_internal_id(dna_case_id)
                if dry_run:
                    continue
                self.scout_api.upload_rna_fraser_outrider(
                    fraser_file_path=rna_fraser.full_path,
                    outrider_file_path=rna_outrider.full_path,
                    case_id=dna_case_id,
                    customer_case_name=customer_case.name,
                    cust_id=cust_id,
                )
        for upload_statement in self.get_rna_fraser_outrider_upload_summary(rna_dna_collections):
            LOG.info(upload_statement)
        LOG.info("Upload RNA fraser and outrider file finished!")

    def upload_rna_genome_build_to_scout(
        self,
        dry_run: bool,
        rna_case: str,
        rna_dna_collections: list[RNADNACollection],
        cust_id: str,
    ) -> None:
        """Upload RNA genome built for a RNA/DNA case to Scout."""
        status_db: Store = self.status_db
        for rna_dna_collection in rna_dna_collections:
            dna_sample_name: str = rna_dna_collection.dna_sample_name
            rna_genome_build = genome_to_scout_format(
                GenomeVersion(get_genome_build(case=rna_case))
            )
            for dna_case_id in rna_dna_collection.dna_case_ids:
                LOG.info(
                    f"Uploading RNA genome built for sample {dna_sample_name} "
                    f"in case {dna_case_id} in Scout."
                )
                customer_case = status_db.get_case_by_internal_id(dna_case_id)

                if dry_run:
                    continue

                self.scout_api.upload_rna_genome_build(
                    case_id=dna_case_id,
                    customer_case_name=customer_case.name,
                    cust_id=cust_id,
                    rna_genome_build=rna_genome_build,
                )

        for upload_statement in self.get_rna_genome_build_upload_summary(rna_dna_collections):
            LOG.info(upload_statement)
        LOG.info("Upload RNA fraser and outrider file finished!")

    def load_rna_variant_outlier_to_scout(
        self, dry_run: bool, rna_dna_collections: list[RNADNACollection]
    ) -> None:
        """Upload RNA genome built for a RNA/DNA case to Scout."""
        for rna_dna_collection in rna_dna_collections:
            dna_sample_name: str = rna_dna_collection.dna_sample_name
            for dna_case_id in rna_dna_collection.dna_case_ids:
                LOG.info(
                    f"Loading RNA variants for sample {dna_sample_name} "
                    f"in case {dna_case_id} in Scout."
                )

                if dry_run:
                    continue

                self.scout_api.load_variant_outlier(
                    case_id=dna_case_id,
                )

        for upload_statement in self.get_variant_load_summary(rna_dna_collections):
            LOG.info(upload_statement)
        LOG.info("Load RNA variants finished!")

    def upload_splice_junctions_bed_to_scout(self, dry_run: bool, case_id: str) -> None:
        """Upload splice_junctions_bed file for a case to Scout."""

        status_db: Store = self.status_db
        rna_case: Case = status_db.get_case_by_internal_id(case_id)

        rna_dna_collections: list[RNADNACollection] = self.create_rna_dna_collections(rna_case)
        for rna_dna_collection in rna_dna_collections:
            rna_sample_internal_id: str = rna_dna_collection.rna_sample_internal_id
            dna_sample_name: str = rna_dna_collection.dna_sample_name
            splice_junctions_bed: File | None = self.get_splice_junctions_bed(
                case_id=case_id, sample_id=rna_sample_internal_id
            )

            if not splice_junctions_bed:
                raise FileNotFoundError(
                    f"No splice junctions bed file was found in Housekeeper for {rna_sample_internal_id}."
                )

            LOG.debug(f"Splice junctions bed file {splice_junctions_bed.path} found")
            for dna_case_id in rna_dna_collection.dna_case_ids:
                LOG.info(
                    f"Uploading splice junctions bed file for sample {dna_sample_name} "
                    f"in case {dna_case_id} in Scout."
                )

                if dry_run:
                    continue

                self.scout_api.upload_splice_junctions_bed(
                    file_path=splice_junctions_bed.full_path,
                    case_id=dna_case_id,
                    customer_sample_id=dna_sample_name,
                )
        for upload_statement in self.get_rna_splice_junctions_upload_summary(rna_dna_collections):
            LOG.info(upload_statement)
        LOG.info("Upload splice junctions bed file finished!")

    @staticmethod
    def get_rna_alignment_file_upload_summary(
        rna_dna_collections: list[RNADNACollection],
    ) -> list[str]:
        upload_summary: list[str] = []
        for rna_dna_collection in rna_dna_collections:
            upload_summary.extend(
                f"Uploaded RNA alignment CRAM file for sample {rna_dna_collection.dna_sample_name} in case {dna_case}"
                for dna_case in rna_dna_collection.dna_case_ids
            )
        return upload_summary

    @staticmethod
    def get_rna_splice_junctions_upload_summary(
        rna_dna_collections: list[RNADNACollection],
    ) -> list[str]:
        upload_summary: list[str] = []
        for rna_dna_collection in rna_dna_collections:
            upload_summary.extend(
                f"Uploaded splice junctions bed file for sample {rna_dna_collection.dna_sample_name} in case {dna_case}."
                for dna_case in rna_dna_collection.dna_case_ids
            )
        return upload_summary

    @staticmethod
    def get_rna_bigwig_coverage_upload_summary(
        rna_dna_collections: list[RNADNACollection],
    ) -> list[str]:
        upload_summary: list[str] = []
        for rna_dna_collection in rna_dna_collections:
            upload_summary.extend(
                f"Uploaded bigwig coverage file for sample {rna_dna_collection.dna_sample_name} in case {dna_case}."
                for dna_case in rna_dna_collection.dna_case_ids
            )
        return upload_summary

    @staticmethod
    def get_rna_fraser_outrider_upload_summary(
        rna_dna_collections: list[RNADNACollection],
    ) -> list[str]:
        upload_summary: list[str] = []
        for rna_dna_collection in rna_dna_collections:
            upload_summary.extend(
                f"Uploaded fraser and outrider files for sample {rna_dna_collection.dna_sample_name} in case {dna_case}."
                for dna_case in rna_dna_collection.dna_case_ids
            )
        return upload_summary

    @staticmethod
    def get_rna_genome_build_upload_summary(
        rna_dna_collections: list[RNADNACollection],
    ) -> list[str]:
        upload_summary: list[str] = []
        for rna_dna_collection in rna_dna_collections:
            upload_summary.extend(
                f"Uploaded RNA genome build file for sample {rna_dna_collection.dna_sample_name} in case {dna_case}."
                for dna_case in rna_dna_collection.dna_case_ids
            )
        return upload_summary

    @staticmethod
    def get_variant_load_summary(
        rna_dna_collections: list[RNADNACollection],
    ) -> list[str]:
        upload_summary: list[str] = []
        for rna_dna_collection in rna_dna_collections:
            upload_summary.extend(
                f"Loaded variant outlier for sample {rna_dna_collection.dna_sample_name} in case {dna_case}."
                for dna_case in rna_dna_collection.dna_case_ids
            )
        return upload_summary

    def upload_rna_junctions_to_scout(self, dry_run: bool, case_id: str) -> None:
        """Upload RNA junctions splice files to Scout."""
        self.upload_splice_junctions_bed_to_scout(dry_run=dry_run, case_id=case_id)
        self.upload_rna_coverage_bigwig_to_scout(case_id=case_id, dry_run=dry_run)

    def upload_rna_omics_to_scout(self, dry_run: bool, case_id: str) -> None:
        """Upload RNA omics files to Scout."""
        status_db: Store = self.status_db
        rna_case = status_db.get_case_by_internal_id(case_id)
        rna_dna_collections: list[RNADNACollection] = self.create_rna_dna_collections(rna_case)
        cust_id: str = rna_case.customer.internal_id
        self.upload_omics_sample_id_to_scout(
            dry_run=dry_run, rna_dna_collections=rna_dna_collections
        )
        self.upload_rna_fraser_outrider_to_scout(
            dry_run=dry_run,
            case_id=case_id,
            rna_dna_collections=rna_dna_collections,
            cust_id=cust_id,
        )
        self.upload_rna_genome_build_to_scout(
            dry_run=dry_run,
            rna_case=rna_case,
            rna_dna_collections=rna_dna_collections,
            cust_id=cust_id,
        )
        self.load_rna_variant_outlier_to_scout(
            dry_run=dry_run, rna_dna_collections=rna_dna_collections
        )

    def get_config_builder(self, analysis, hk_version) -> ScoutConfigBuilder:
        config_builders = {
            Workflow.BALSAMIC: BalsamicConfigBuilder(
                hk_version_obj=hk_version,
                analysis_obj=analysis,
                lims_api=self.lims,
            ),
            Workflow.BALSAMIC_UMI: BalsamicUmiConfigBuilder(
                hk_version_obj=hk_version,
                analysis_obj=analysis,
                lims_api=self.lims,
            ),
            Workflow.MIP_DNA: MipConfigBuilder(
                hk_version_obj=hk_version,
                analysis_obj=analysis,
                mip_analysis_api=self.analysis_api,
                lims_api=self.lims,
                madeline_api=self.madeline_api,
            ),
            Workflow.MIP_RNA: MipConfigBuilder(
                hk_version_obj=hk_version,
                analysis_obj=analysis,
                mip_analysis_api=self.analysis_api,
                lims_api=self.lims,
                madeline_api=self.madeline_api,
            ),
            Workflow.RAREDISEASE: RarediseaseConfigBuilder(
                hk_version_obj=hk_version,
                analysis_obj=analysis,
                raredisease_analysis_api=self.raredisease_analysis_api,
                lims_api=self.lims,
                madeline_api=self.madeline_api,
            ),
            Workflow.RNAFUSION: RnafusionConfigBuilder(
                hk_version_obj=hk_version,
                analysis_obj=analysis,
                lims_api=self.lims,
            ),
        }

        return config_builders[analysis.workflow]

    def create_rna_dna_collections(self, rna_case: Case) -> list[RNADNACollection]:
        return [self.create_rna_dna_collection(link.sample) for link in rna_case.links]

    def create_rna_dna_collection(self, rna_sample: Sample) -> RNADNACollection:
        """Creates a collection containing the given RNA sample id, its related DNA sample name, and
        a list of ids for the DNA cases connected to the DNA sample."""
        if not rna_sample.subject_id:
            raise CgDataError(
                f"Failed to link RNA sample {rna_sample.internal_id} to DNA samples - subject_id field is empty."
            )

        collaborators: set[Customer] = rna_sample.customer.collaborators
        subject_id_samples: list[Sample] = (
            self.status_db.get_samples_by_customer_ids_and_subject_id_and_is_tumour(
                customer_ids=[customer.id for customer in collaborators],
                subject_id=rna_sample.subject_id,
                is_tumour=rna_sample.is_tumour,
            )
        )

        subject_id_dna_samples: list[Sample] = self._get_application_prep_category(
            subject_id_samples
        )

        if len(subject_id_dna_samples) != 1:
            raise CgDataError(
                f"Failed to upload files for RNA case: unexpected number of DNA sample matches for subject_id: "
                f"{rna_sample.subject_id}. Number of matches: {len(subject_id_dna_samples)} "
            )
        dna_sample: Sample = subject_id_dna_samples[0]
        dna_cases: list[str] = self._dna_cases_related_to_dna_sample(
            dna_sample=dna_sample, collaborators=collaborators
        )
        return RNADNACollection(
            rna_sample_internal_id=rna_sample.internal_id,
            dna_sample_name=dna_sample.name,
            dna_case_ids=dna_cases,
        )

    def _dna_cases_related_to_dna_sample(
        self, dna_sample: Sample, collaborators: set[Customer]
    ) -> list[str]:
        """Maps a list of uploaded DNA cases linked to DNA sample."""
        potential_cases_related_to_dna_sample: list[Case] = [link.case for link in dna_sample.links]
        return self.filter_cases_related_to_dna_sample(
            list_of_dna_cases=potential_cases_related_to_dna_sample, collaborators=collaborators
        )

    @staticmethod
    def filter_cases_related_to_dna_sample(
        list_of_dna_cases: list[Case], collaborators: set[Customer]
    ) -> list[str]:
        """Filters the given list of DNA samples and returns a subset of uploaded cases ordered by customers in the
        specified list of collaborators and within the correct workflow."""
        filtered_dna_cases: list[str] = []
        for case in list_of_dna_cases:
            if (
                case.data_analysis
                in [
                    Workflow.BALSAMIC,
                    Workflow.BALSAMIC_UMI,
                    Workflow.MIP_DNA,
                    Workflow.RAREDISEASE,
                ]
                and case.customer in collaborators
            ):
                if not case.is_uploaded:
                    LOG.warning(
                        f"Cannot upload RNA report to DNA case {case.internal_id}. DNA case is not uploaded."
                    )
                filtered_dna_cases.append(case.internal_id)
        return filtered_dna_cases

    @staticmethod
    def _get_application_prep_category(
        subject_id_samples: list[Sample],
    ) -> list[Sample | None]:
        """Filter a models Sample list, returning DNA samples selected on their preparation category."""
        subject_id_dna_samples: list[Sample | None] = [
            sample
            for sample in subject_id_samples
            if sample.prep_category
            in [
                PrepCategory.WHOLE_GENOME_SEQUENCING.value,
                PrepCategory.TARGETED_GENOME_SEQUENCING.value,
                PrepCategory.WHOLE_EXOME_SEQUENCING.value,
            ]
        ]

        return subject_id_dna_samples

    def get_related_uploaded_dna_cases(self, rna_case_id: str) -> set[str]:
        """Returns all uploaded DNA cases related to the specified RNA case."""
        unique_dna_case_ids: set[str] = self.get_unique_dna_cases_related_to_rna_case(rna_case_id)
        uploaded_dna_cases: set[str] = set()
        for dna_case_id in unique_dna_case_ids:
            if self.status_db.get_case_by_internal_id(dna_case_id).is_uploaded:
                uploaded_dna_cases.add(dna_case_id)
            else:
                LOG.warning(f"Related DNA case {dna_case_id} has not been completed.")
        return uploaded_dna_cases



import logging

from housekeeper.store.models import Version

from cg.apps.lims import LimsAPI
from cg.apps.madeline.api import MadelineAPI
from cg.constants.housekeeper_tags import HK_DELIVERY_REPORT_TAG
from cg.constants.scout import (
    RAREDISEASE_CASE_TAGS,
    RAREDISEASE_SAMPLE_TAGS,
    GenomeBuild,
    UploadTrack,
)
from cg.meta.upload.scout.hk_tags import CaseTags, SampleTags
from cg.meta.upload.scout.scout_config_builder import ScoutConfigBuilder
from cg.meta.workflow.raredisease import RarediseaseAnalysisAPI
from cg.models.scout.scout_load_config import (
    CaseImages,
    CustomImages,
    Eklipse,
    RarediseaseLoadConfig,
    ScoutRarediseaseIndividual,
)
from cg.store.models import Analysis

LOG = logging.getLogger(__name__)


class RarediseaseConfigBuilder(ScoutConfigBuilder):
    def __init__(
        self,
        hk_version_obj: Version,
        analysis_obj: Analysis,
        raredisease_analysis_api: RarediseaseAnalysisAPI,
        lims_api: LimsAPI,
        madeline_api: MadelineAPI,
    ):
        super().__init__(
            hk_version_obj=hk_version_obj,
            analysis_obj=analysis_obj,
            lims_api=lims_api,
        )
        self.case_tags: CaseTags = CaseTags(**RAREDISEASE_CASE_TAGS)
        self.sample_tags: SampleTags = SampleTags(**RAREDISEASE_SAMPLE_TAGS)
        self.raredisease_analysis_api: RarediseaseAnalysisAPI = raredisease_analysis_api
        self.lims_api: LimsAPI = lims_api
        self.madeline_api: MadelineAPI = madeline_api

    def build_load_config(self) -> RarediseaseLoadConfig:
        """Create a RAREDISEASE specific load config for uploading analysis to Scout."""
        LOG.info("Build load config for RAREDISEASE case")
        load_config: RarediseaseLoadConfig = RarediseaseLoadConfig(
            track=UploadTrack.RARE_DISEASE.value,
            delivery_report=self.get_file_from_hk({HK_DELIVERY_REPORT_TAG}),
        )
        self.add_common_info_to_load_config(load_config)
        load_config.gene_panels = self.raredisease_analysis_api.get_aggregated_panels(
            customer_id=self.analysis_obj.case.customer.internal_id,
            default_panels=set(self.analysis_obj.case.panels),
        )
        self.include_case_files(load_config)
        self.get_sample_information(load_config)
        self.include_pedigree_picture(load_config)
        self.load_custom_image_sample(load_config)
        load_config.human_genome_build = GenomeBuild.hg19
        return load_config

    def load_custom_image_sample(self, load_config: RarediseaseLoadConfig) -> None:
        """Build custom images config."""
        LOG.info("Adding custom images")

        eklipse_images: list = []
        for sample in self.analysis_obj.case.samples:
            sample_id: str = sample.internal_id
            eklipse_image_path = self.get_file_from_hk(
                hk_tags=set(self.sample_tags.eklipse_path).union({sample_id})
            )
            if eklipse_image_path:
                eklipse_image = Eklipse(
                    title=sample_id,
                    path=eklipse_image_path,
                    description="eKLIPse MT images",
                    width="800",
                    height="800",
                )
                eklipse_images.append(eklipse_image)
        if eklipse_images:
            case_images = CaseImages(eKLIPse=eklipse_images)
            config_custom_images = CustomImages(case_images=case_images)
            load_config.custom_images = config_custom_images

    def include_case_files(self, load_config: RarediseaseLoadConfig) -> None:
        """Include case level files for mip case."""
        LOG.info("Including RAREDISEASE specific case level files")
        for scout_key in RAREDISEASE_CASE_TAGS.keys():
            self._include_case_file(load_config, scout_key)

    def _include_case_file(self, load_config: RarediseaseLoadConfig, scout_key: str) -> None:
        """Include the file path associated to a scout configuration parameter if the corresponding housekeeper tags
        are found. Otherwise return None."""
        file_path = self.get_file_from_hk(getattr(self.case_tags, scout_key))
        setattr(load_config, scout_key, file_path)

    def include_sample_files(self, config_sample: ScoutRarediseaseIndividual) -> None:
        """Include sample level files that are optional for mip samples."""
        LOG.info("Including RAREDISEASE specific sample level files")
        sample_id: str = config_sample.sample_id
        config_sample.vcf2cytosure = self.get_sample_file(
            hk_tags=self.sample_tags.vcf2cytosure, sample_id=sample_id
        )
        config_sample.mt_bam = self.get_sample_file(
            hk_tags=self.sample_tags.mt_bam, sample_id=sample_id
        )
        config_sample.chromograph_images.autozygous = self.remove_chromosome_substring(
            file_path=self.get_sample_file(
                hk_tags=self.sample_tags.chromograph_autozyg, sample_id=sample_id
            )
        )
        config_sample.chromograph_images.coverage = self.remove_chromosome_substring(
            file_path=self.get_sample_file(
                hk_tags=self.sample_tags.chromograph_coverage, sample_id=sample_id
            )
        )
        config_sample.chromograph_images.upd_regions = self.remove_chromosome_substring(
            file_path=self.get_sample_file(
                hk_tags=self.sample_tags.chromograph_regions, sample_id=sample_id
            )
        )
        config_sample.chromograph_images.upd_sites = self.remove_chromosome_substring(
            file_path=self.get_sample_file(
                hk_tags=self.sample_tags.chromograph_sites, sample_id=sample_id
            )
        )
        config_sample.reviewer.alignment = self.get_sample_file(
            hk_tags=self.sample_tags.reviewer_alignment, sample_id=sample_id
        )
        config_sample.reviewer.alignment_index = self.get_sample_file(
            hk_tags=self.sample_tags.reviewer_alignment_index, sample_id=sample_id
        )
        config_sample.reviewer.vcf = self.get_sample_file(
            hk_tags=self.sample_tags.reviewer_vcf, sample_id=sample_id
        )
        config_sample.reviewer.catalog = self.get_file_from_hk(hk_tags=self.case_tags.str_catalog)
        config_sample.mitodel_file = self.get_sample_file(
            hk_tags=self.sample_tags.mitodel_file, sample_id=sample_id
        )



from pathlib import Path

from pydantic import BaseModel, ValidationInfo, conlist, field_validator

from cg.exc import NfSampleSheetError


class WorkflowParameters(BaseModel):
    input: Path
    outdir: Path


class NfCommandArgs(BaseModel):
    """Model for arguments and options supported."""

    log: str | Path | None = None
    resume: bool | None = None
    profile: str | None = None
    stub_run: bool | None = None
    config: str | Path | None = None
    name: str | None = None
    revision: str | None = None
    wait: str | None = None
    id: str | None = None
    with_tower: bool | None = None
    use_nextflow: bool | None = None
    compute_env: str | None = None
    work_dir: str | Path | None = None
    params_file: str | Path | None = None


class NextflowSampleSheetEntry(BaseModel):
    """Nextflow sample sheet model.

    Attributes:
        name: sample name, or case id
        fastq_forward_read_paths: list of all fastq read1 file paths corresponding to sample
        fastq_reverse_read_paths: list of all fastq read2 file paths corresponding to sample
    """

    name: str
    fastq_forward_read_paths: conlist(Path, min_length=1)
    fastq_reverse_read_paths: conlist(Path, min_length=1)

    @field_validator("fastq_reverse_read_paths")
    @classmethod
    def validate_complete_fastq_file_pairs(
        cls, fastq_reverse: list[str], info: ValidationInfo
    ) -> list[str]:
        """Verify that the number of fastq forward files is the same as for the reverse."""
        if len(fastq_reverse) != len(info.data.get("fastq_forward_read_paths")):
            raise NfSampleSheetError("Fastq file length for forward and reverse do not match")
        return fastq_reverse

    @field_validator("fastq_forward_read_paths", "fastq_reverse_read_paths")
    @classmethod
    def fastq_files_exist(cls, fastq_paths: list[str]) -> list[str]:
        """Verify that fastq files exist."""
        for fastq_path in fastq_paths:
            if not fastq_path.is_file():
                raise NfSampleSheetError(f"Fastq file does not exist: {str(fastq_path)}")
        return fastq_paths


class FileDeliverable(BaseModel):
    """Specification for a general deliverables file."""

    id: str
    format: str
    path: str
    path_index: str | None = None
    step: str
    tag: str

    @field_validator("path", "path_index", mode="before")
    @classmethod
    def set_path_as_string(cls, file_path: str | Path) -> str | None:
        return str(Path(file_path)) if file_path else None


class WorkflowDeliverables(BaseModel):
    """Specification for workflow deliverables."""

    files: list[FileDeliverable]



"""Module to read or write txt files."""

from pathlib import Path

from cg.constants.symbols import EMPTY_STRING


def read_txt(file_path: Path, read_to_string: bool = False) -> list[str] | str:
    """Read content in a TXT file."""
    with open(file_path, "r") as file:
        if read_to_string:
            return file.read()
        return list(file)


def write_txt(content: list[str] | str, file_path: Path) -> None:
    """Write content to a text file."""
    with open(file_path, "w") as file:
        if isinstance(content, list):
            file.writelines(content)
        else:
            file.write(content)


def concat_txt(
    file_paths: list[Path | str],
    str_content: list[str] | None = None,
) -> str | None:
    """Concatenate the content of several files and eventual string content."""
    content: str = EMPTY_STRING
    if str_content:
        for txt in str_content:
            content += f"{txt}\n"
    for file_path in file_paths:
        file_content: str = read_txt(file_path, read_to_string=True)
        content += f"{file_content}\n"
    return content



import os
from pathlib import Path
from cg.services.deliver_files.file_fetcher.models import SampleFile
from cg.services.deliver_files.file_formatter.models import FormattedFile


class SampleFileFormatter:
    """
    Format the sample files to deliver.
    Used for all workflows except Microsalt and Mutant.
    """

    def format_files(
        self, moved_files: list[SampleFile], ticket_dir_path: Path
    ) -> list[FormattedFile]:
        """Format the sample files to deliver and return the formatted files."""
        sample_names: set[str] = self._get_sample_names(moved_files)
        self._create_sample_folders(ticket_dir_path=ticket_dir_path, sample_names=sample_names)
        return self._format_sample_files(moved_files)

    @staticmethod
    def _get_sample_names(sample_files: list[SampleFile]) -> set[str]:
        return set(sample_file.sample_name for sample_file in sample_files)

    @staticmethod
    def _create_sample_folders(ticket_dir_path: Path, sample_names: set[str]):
        for sample_name in sample_names:
            sample_dir_path = Path(ticket_dir_path, sample_name)
            sample_dir_path.mkdir(exist_ok=True)

    def _format_sample_files(self, sample_files: list[SampleFile]) -> list[FormattedFile]:
        formatted_files: list[FormattedFile] = self._get_formatted_files(sample_files)
        for formatted_file in formatted_files:
            os.rename(src=formatted_file.original_path, dst=formatted_file.formatted_path)
        return formatted_files

    @staticmethod
    def _get_formatted_files(sample_files: list[SampleFile]) -> list[FormattedFile]:
        """
        Returns formatted files:
        1. Adds a folder with sample name to the path of the sample files.
        2. Replaces sample id by sample name.
        """
        formatted_files: list[FormattedFile] = []
        for sample_file in sample_files:
            replaced_sample_file_name: str = sample_file.file_path.name.replace(
                sample_file.sample_id, sample_file.sample_name
            )
            formatted_file_path = Path(
                sample_file.file_path.parent, sample_file.sample_name, replaced_sample_file_name
            )
            formatted_files.append(
                FormattedFile(
                    original_path=sample_file.file_path, formatted_path=formatted_file_path
                )
            )
        return formatted_files



import pytest

from cg.constants import DataDelivery, Workflow
from cg.models.orders.order import OrderIn
from cg.models.orders.sample_base import SexEnum
from cg.models.orders.samples import PacBioSample
from cg.services.orders.validate_order_services.validate_pacbio_order import (
    ValidatePacbioOrderService,
)
from cg.store.store import Store


@pytest.fixture
def pacbio_sample() -> PacBioSample:
    return PacBioSample(
        application="WGSPCFC060",
        data_analysis=Workflow.RAW_DATA,
        data_delivery=DataDelivery.NO_DELIVERY,
        name="PacbioSample",
        sex=SexEnum.unknown,
        tumour=False,
        volume="50",
        buffer="buffer",
        source="source",
        subject_id="subject_id",
        container="Tube",
    )


@pytest.fixture
def pacbio_order(pacbio_sample: PacBioSample) -> OrderIn:
    return OrderIn(
        customer="cust000",
        name="PacbioOrder",
        samples=[pacbio_sample],
    )


@pytest.fixture
def validate_pacbio_order_service(sample_store: Store) -> ValidatePacbioOrderService:
    return ValidatePacbioOrderService(sample_store)



import pytest

from cg.apps.mip.confighandler import ConfigHandler
from cg.exc import PedigreeConfigError


def test_validate_config(valid_config):
    """Test to validate a correct config"""
    # GIVEN a correct config

    # WHEN validating the config
    errors = ConfigHandler.validate_config(valid_config)

    # THEN assert the errors is a dict
    assert isinstance(errors, dict)

    # THEN assert there were no errors
    assert errors == {}


def test_validate_config_invalid_analysis_type(invalid_config_analysis_type):
    """Test to validate a invalid config with wrong analysis type"""
    # GIVEN invalid config, where nalaysis type is not correct

    # WHEN validating the config
    with pytest.raises(PedigreeConfigError):
        # THEN assert a config error is raised
        ConfigHandler.validate_config(invalid_config_analysis_type)


def test_validate_config_unknown_field(invalid_config_unknown_field):
    """Test to validate a config with a unspecified field.

    This should work since we allow unspecified fields
    """
    # GIVEN a correct config with a extra field

    # WHEN validating the config
    errors = ConfigHandler.validate_config(invalid_config_unknown_field)

    # THEN assert the errors is a dict
    assert isinstance(errors, dict)

    # THEN assert there where one error that could pass
    assert len(errors) == 1


def test_validate_config_unknown_field_and_missing_sample_id(
    invalid_config_unknown_field_sample_id,
):
    """Test to validate a config with a unspecified field and a missing mandatory field.
    This should raise exception since we do not allow missing mandatory fields
    """
    # GIVEN a config with missing sample_id and an extra field

    # WHEN validating the config
    with pytest.raises(PedigreeConfigError):
        # THEN assert a config error is raised
        ConfigHandler.validate_config(invalid_config_unknown_field_sample_id)


def test_validate_config_unknown_field_and_invalid_analysis_type(
    invalid_config_unknown_field_analysis_type,
):
    """Test to validate a config with a unspecified field and a invalid mandatory field.
    This should raise exception since we do not allow wrong analysis types
    """
    # GIVEN a config with wrong analysis type and an extra field

    # WHEN validating the config
    with pytest.raises(PedigreeConfigError):
        # THEN assert a config error is raised
        ConfigHandler.validate_config(invalid_config_unknown_field_analysis_type)



"""Module to test Janus API client."""

import pytest
import requests
from pytest_mock import MockFixture

from cg.clients.janus.api import JanusAPIClient
from cg.clients.janus.dto.create_qc_metrics_request import CreateQCMetricsRequest
from cg.clients.janus.exceptions import JanusClientError


def test_qc_metrics_successful(
    mocker: MockFixture,
    janus_client: JanusAPIClient,
    collect_qc_request_balsamic_wgs: CreateQCMetricsRequest,
    janus_response: dict,
    mock_post_request_ok: MockFixture,
):
    # GIVEN a mocked response from the requests.post method that is successful
    mocked_post = mocker.patch.object(requests, "post")
    mocked_post.return_value = mock_post_request_ok

    # WHEN retrieving the qc metrics
    jobs_response: dict = janus_client.qc_metrics(collect_qc_request_balsamic_wgs)

    # THEN the qc metrics are deserialized without error
    assert jobs_response == janus_response
    mocked_post.assert_called_once_with(
        f"{janus_client.host}/collect_qc",
        data=collect_qc_request_balsamic_wgs.model_dump_json(),
        verify=True,
    )


def test_qc_metrics_not_successful(
    mocker: MockFixture,
    janus_client: JanusAPIClient,
    collect_qc_request_balsamic_wgs: CreateQCMetricsRequest,
    janus_response: dict,
    mock_post_request_not_found: MockFixture,
):
    # GIVEN a mocked response from the requests.post method that is not successful
    mocked_post = mocker.patch.object(requests, "post")
    mocked_post.return_value = mock_post_request_not_found

    # WHEN retrieving the qc metrics

    # THEN the request raises an error
    with pytest.raises(JanusClientError):
        janus_client.qc_metrics(collect_qc_request_balsamic_wgs)



import logging

from cg.apps.lims import LimsAPI
from cg.models.lims.sample import LimsSample
from cg.models.orders.order import OrderIn
from cg.models.orders.samples import OrderInSample

LOG = logging.getLogger(__name__)


class OrderLimsService:

    def __init__(self, lims_api: LimsAPI):
        self.lims_api = lims_api

    @staticmethod
    def _build_lims_sample(customer: str, samples: list[OrderInSample]) -> list[LimsSample]:
        """Convert order input to lims interface input."""
        samples_lims = []
        for sample in samples:
            dict_sample = sample.__dict__
            LOG.debug(f"{sample.name}: prepare LIMS input")
            dict_sample["customer"] = customer
            lims_sample: LimsSample = LimsSample.parse_obj(dict_sample)
            samples_lims.append(lims_sample)
        return samples_lims

    def process_lims(self, lims_order: OrderIn, new_samples: list[OrderInSample]):
        """Process samples to add them to LIMS."""
        samples_lims: list[LimsSample] = self._build_lims_sample(
            lims_order.customer, samples=new_samples
        )
        project_name: str = lims_order.ticket or lims_order.name
        # Create new lims project
        project_data = self.lims_api.submit_project(
            project_name, [lims_sample.dict() for lims_sample in samples_lims]
        )
        lims_map = self.lims_api.get_samples(projectlimsid=project_data["id"], map_ids=True)
        return project_data, lims_map



"""Module for PacBio fixtures returning strings."""

import pytest


@pytest.fixture
def pac_bio_smrt_cell_name() -> str:
    return "1_B01"


@pytest.fixture
def pacbio_barcoded_smrt_cell_name() -> str:
    return "1_C01"


@pytest.fixture
def pac_bio_test_run_name() -> str:
    """Return the name of a PacBio SMRT cell."""
    return "r84202_20240522_133539"


@pytest.fixture
def pacbio_barcoded_run_name() -> str:
    return "r84202_20240913_121403"


@pytest.fixture
def pacbio_sequencing_run_name(pac_bio_test_run_name: str, pac_bio_smrt_cell_name: str) -> str:
    """Return the name of a PacBio SMRT cell."""
    return f"{pac_bio_test_run_name}/{pac_bio_smrt_cell_name}"


@pytest.fixture
def pacbio_barcoded_sequencing_run_name(
    pacbio_barcoded_run_name: str, pacbio_barcoded_smrt_cell_name: str
) -> str:
    """Return the name of a PacBio SMRT cell."""
    return f"{pacbio_barcoded_run_name}/{pacbio_barcoded_smrt_cell_name}"


@pytest.fixture
def pacbio_processed_sequencing_run_name(
    pacbio_barcoded_run_name: str, pac_bio_smrt_cell_name: str
) -> str:
    """Return the name of a PacBio SMRT cell."""
    return f"{pacbio_barcoded_run_name}/{pac_bio_smrt_cell_name}"


@pytest.fixture
def pacbio_run_names(
    pacbio_sequencing_run_name: str,
    pacbio_barcoded_sequencing_run_name: str,
    pacbio_processed_sequencing_run_name: str,
) -> set[str]:
    return {
        pacbio_sequencing_run_name,
        pacbio_barcoded_sequencing_run_name,
        pacbio_processed_sequencing_run_name,
    }


@pytest.fixture
def pacbio_barcoded_1_c01_cell_full_name() -> str:
    """Return the full name of a PacBio SMRT cell."""
    return "m84202_240913_162115_s3"


@pytest.fixture
def barcoded_smrt_cell_internal_id() -> str:
    return "EA114368"


@pytest.fixture
def pacbio_barcoded_sample_internal_id() -> str:
    return "Bio Sample 4"


@pytest.fixture
def pacbio_unassigned_sample_internal_id() -> str:
    return "No Name"



from pathlib import Path

import pytest
from cg.constants.constants import ReadDirection

from cg.services.fastq_concatenation_service.fastq_concatenation_service import (
    FastqConcatenationService,
)
from cg.services.fastq_concatenation_service.utils import generate_concatenated_fastq_delivery_path


def test_empty_directory(fastq_file_service: FastqConcatenationService, tmp_path):
    # GIVEN an empty directory

    # GIVEN output files
    forward_output_path = Path(tmp_path, "forward.fastq.gz")
    reverse_output_path = Path(tmp_path, "reverse.fastq.gz")

    # WHEN concatenating the reads
    fastq_file_service.concatenate(
        fastq_directory=tmp_path,
        forward_output_path=forward_output_path,
        reverse_output_path=reverse_output_path,
    )

    # THEN the output files should not exist
    assert not forward_output_path.exists()
    assert not reverse_output_path.exists()


def test_concatenate(fastq_file_service: FastqConcatenationService, fastqs_dir: Path):
    # GIVEN a directory with forward and reverse reads

    # GIVEN output files for the concatenated reads
    forward_output_path = Path(fastqs_dir, "forward.fastq.gz")
    reverse_output_path = Path(fastqs_dir, "reverse.fastq.gz")

    # WHEN concatenating the reads
    fastq_file_service.concatenate(
        fastq_directory=fastqs_dir,
        forward_output_path=forward_output_path,
        reverse_output_path=reverse_output_path,
        remove_raw=True,
    )

    # THEN the output files should exist
    assert forward_output_path.exists()
    assert reverse_output_path.exists()

    # THEN the concatenated forward reads only contain forward reads
    assert "forward" in forward_output_path.read_text()
    assert "reverse" not in forward_output_path.read_text()

    # THEN the concatenated reverse reads only contain reverse reads
    assert "reverse" in reverse_output_path.read_text()
    assert "forward" not in reverse_output_path.read_text()


def test_concatenate_when_output_exists(
    fastq_file_service: FastqConcatenationService, fastqs_dir: Path
):
    # GIVEN a directory with forward and reverse reads
    existing_fastq_files = list(fastqs_dir.iterdir())
    existing_forward: Path = existing_fastq_files[0]

    # GIVEN that the forward output file already exists
    forward_output_path = existing_forward
    reverse_output_path = Path(fastqs_dir, "reverse.fastq.gz")

    # WHEN concatenating the reads
    fastq_file_service.concatenate(
        fastq_directory=fastqs_dir,
        forward_output_path=forward_output_path,
        reverse_output_path=reverse_output_path,
        remove_raw=True,
    )

    # THEN the output files should exist
    assert forward_output_path.exists()
    assert reverse_output_path.exists()

    # THEN the concatenated forward reads only contain forward reads
    assert "forward" in forward_output_path.read_text()
    assert "reverse" not in forward_output_path.read_text()

    # THEN the concatenated reverse reads only contain reverse reads
    assert "reverse" in reverse_output_path.read_text()
    assert "forward" not in reverse_output_path.read_text()


def test_concatenate_missing_reverse(
    fastq_file_service: FastqConcatenationService, fastqs_forward: Path, tmp_path
):
    # GIVEN a directory with forward reads only

    # GIVEN output files for the concatenated reads
    forward_output_path = Path(tmp_path, "forward.fastq.gz")
    reverse_output_path = Path(tmp_path, "reverse.fastq.gz")

    # WHEN concatenating the reads
    fastq_file_service.concatenate(
        fastq_directory=fastqs_forward,
        forward_output_path=forward_output_path,
        reverse_output_path=reverse_output_path,
    )

    # THEN forward reads should be concatenated
    assert forward_output_path.exists()

    # THEN reverse reads should not exist
    assert not reverse_output_path.exists()


@pytest.mark.parametrize(
    "fastq_directory, sample_name, direction, expected_output_path",
    [
        (
            Path("/path/to/fastq_directory"),
            "sample1",
            ReadDirection.FORWARD,
            Path("/path/to/fastq_directory/sample1_1.fastq.gz"),
        ),
        (
            Path("/path/to/fastq_directory"),
            "sample1",
            ReadDirection.REVERSE,
            Path("/path/to/fastq_directory/sample1_2.fastq.gz"),
        ),
    ],
    ids=["forward", "reverse"],
)
def test_generate_concatenated_fastq_delivery_path(
    fastq_directory: Path, sample_name: str, direction: int, expected_output_path: Path
):
    # GIVEN a fastq direcrory, a sample name and a read direction

    # WHEN generating the concatenated fastq delivery path

    # THEN the output path should be as expected
    assert (
        generate_concatenated_fastq_delivery_path(
            fastq_directory=fastq_directory, sample_name=sample_name, direction=direction
        )
        == expected_output_path
    )



from cg.services.deliver_files.file_fetcher.models import DeliveryFiles
from cg.services.deliver_files.file_filter.sample_service import SampleFileFilter


def test_filter_delivery_files(expected_fastq_delivery_files: DeliveryFiles, sample_id: str):
    """Test to filter delivery files."""

    # GIVEN a delivery files object with multiple sample ids and a filter delivery files service
    filter_service = SampleFileFilter()
    samples_ids: list[str] = [
        sample.sample_id for sample in expected_fastq_delivery_files.sample_files
    ]
    assert len(set(samples_ids)) > 1

    # WHEN filtering the delivery files
    filtered_delivery_files = filter_service.filter_delivery_files(
        expected_fastq_delivery_files, sample_id
    )

    # THEN assert that the delivery files only contains the sample with the given sample id
    for sample_file in filtered_delivery_files.sample_files:
        assert sample_file.sample_id == sample_id



"""Module to test the illumina post processing service."""

from cg.models.run_devices.illumina_run_directory_data import IlluminaRunDirectoryData
from cg.services.illumina.post_processing.post_processing_service import (
    IlluminaPostProcessingService,
)
from cg.store.models import (
    IlluminaFlowCell,
    IlluminaSampleSequencingMetrics,
    IlluminaSequencingRun,
    Sample,
)
from cg.store.store import Store


def test_store_illumina_flow_cell(
    store: Store,
    novaseq_x_demux_runs_flow_cell: IlluminaRunDirectoryData,
    illumina_post_postprocessing_service: IlluminaPostProcessingService,
):
    # GIVEN a flow cell directory data and an Illumina post processing service

    # GIVEN a store without any Illumina flow cells
    assert store._get_query(table=IlluminaFlowCell).count() == 0

    # WHEN storing an Illumina flow cell in the status db
    flow_cell: IlluminaFlowCell = illumina_post_postprocessing_service.store_illumina_flow_cell(
        run_directory_data=novaseq_x_demux_runs_flow_cell
    )

    # THEN assert that the flow cell is created
    assert isinstance(flow_cell, IlluminaFlowCell)
    assert flow_cell.internal_id == novaseq_x_demux_runs_flow_cell.id

    # THEN assert that the flow cell is stored in the status db
    assert store._get_query(table=IlluminaFlowCell).all()


def test_store_illumina_sequencing_run(
    store: Store,
    novaseq_x_demux_runs_flow_cell: IlluminaRunDirectoryData,
    illumina_flow_cell: IlluminaFlowCell,
    illumina_post_postprocessing_service: IlluminaPostProcessingService,
):
    # GIVEN a flow cell directory data and an Illumina post processing service

    # GIVEN a store without any Illumina sequencing runs
    assert store._get_query(table=IlluminaSequencingRun).count() == 0

    # WHEN storing an Illumina sequencing run in the status db
    sequencing_run: IlluminaSequencingRun = (
        illumina_post_postprocessing_service.store_illumina_sequencing_run(
            run_directory_data=novaseq_x_demux_runs_flow_cell, flow_cell=illumina_flow_cell
        )
    )

    # THEN assert that the sequencing run is created
    assert isinstance(sequencing_run, IlluminaSequencingRun)

    # THEN assert that the sequencing run is stored in the status db
    assert store._get_query(table=IlluminaSequencingRun).all()


def test_store_illumina_sample_sequencing_metrics(
    store: Store,
    novaseq_x_demux_runs_flow_cell: IlluminaRunDirectoryData,
    illumina_post_postprocessing_service: IlluminaPostProcessingService,
):
    # GIVEN a store without any Illumina sample sequencing metrics
    assert store._get_query(table=IlluminaSampleSequencingMetrics).count() == 0

    # GIVEN a flow cell directory data with samples and an Illumina post processing service
    flow_cell: IlluminaFlowCell = illumina_post_postprocessing_service.store_illumina_flow_cell(
        novaseq_x_demux_runs_flow_cell
    )
    sequencing_run: IlluminaSequencingRun = (
        illumina_post_postprocessing_service.store_illumina_sequencing_run(
            run_directory_data=novaseq_x_demux_runs_flow_cell, flow_cell=flow_cell
        )
    )

    # WHEN storing the Illumina sample sequencing metrics in the status db
    illumina_post_postprocessing_service.store_illumina_sample_sequencing_metrics(
        run_directory_data=novaseq_x_demux_runs_flow_cell,
        sequencing_run=sequencing_run,
    )

    # THEN the metrics are found in store
    assert store._get_query(table=IlluminaSampleSequencingMetrics).all()


def test_store_illumina_flow_cell_data(
    novaseq_x_demux_runs_flow_cell: IlluminaRunDirectoryData,
    illumina_post_postprocessing_service: IlluminaPostProcessingService,
):
    # GIVEN a flow cell directory data and an Illumina post processing service

    # WHEN storing the flow cell data
    illumina_post_postprocessing_service.store_sequencing_data_in_status_db(
        run_directory_data=novaseq_x_demux_runs_flow_cell
    )

    # THEN assert there is an IlluminaFlowCell in status db
    flow_cell: IlluminaFlowCell = illumina_post_postprocessing_service.status_db._get_query(
        table=IlluminaFlowCell
    ).first()
    assert flow_cell.internal_id == novaseq_x_demux_runs_flow_cell.id

    # THEN assert there is an IlluminaSequencingRun in status db
    sequencing_run: IlluminaSequencingRun = (
        illumina_post_postprocessing_service.status_db._get_query(
            table=IlluminaSequencingRun
        ).first()
    )
    assert sequencing_run

    # THEN illumina sample sequencing metrics are stored in status db
    sample_metrics: list[IlluminaSampleSequencingMetrics] = (
        illumina_post_postprocessing_service.status_db._get_query(
            table=IlluminaSampleSequencingMetrics
        ).all()
    )
    assert sample_metrics

    # THEN the relationship between the flow cell, the sequencing run and sample metrics is correct
    assert sequencing_run.device == flow_cell
    assert sequencing_run.sample_metrics == sample_metrics
    assert flow_cell.instrument_runs == [sequencing_run]
    assert all(sample_metric.instrument_run == sequencing_run for sample_metric in sample_metrics)

    # THEN samples have sample metrics associated to them
    assert all(sample_metric.sample_id for sample_metric in sample_metrics)

    # THEN the samples associated to the stored metrics have their reads and last_sequenced_at updated
    samples: list[Sample] = illumina_post_postprocessing_service.status_db._get_query(Sample).all()
    for sample in samples:
        all_sample_metrics: list[IlluminaSampleSequencingMetrics] = sample.sample_run_metrics
        total_reads_for_sample: int = sum(
            sample_metric.total_reads_in_lane for sample_metric in all_sample_metrics
        )
        assert sample.reads == total_reads_for_sample
        assert sample.last_sequenced_at == sequencing_run.sequencing_completed_at



import pytest

from pathlib import Path

from cg.meta.workflow.mutant.quality_controller.metrics_parser_utils import (
    _get_validated_results_list,
    parse_samples_results,
)
from cg.meta.workflow.mutant.quality_controller.models import (
    SamplePoolAndResults,
    ParsedSampleResults,
    SamplesQualityResults,
)
from cg.meta.workflow.mutant.quality_controller.quality_controller import MutantQualityController
from cg.store.models import Case, Sample
from cg.store.store import Store
from cg.constants.constants import ControlOptions, MutantQC
from tests.store_helpers import StoreHelpers
from tests.mocks.limsmock import LimsSample, LimsUDF, MockLimsAPI


@pytest.fixture
def mutant_store(store: Store, helpers: StoreHelpers) -> Store:
    # Add mutant application and application_version
    application = helpers.add_application(
        store=store, application_tag="VWGDPTR001", target_reads=2000000, percent_reads_guaranteed=1
    )

    # Add cases
    case_qc_pass = helpers.add_case(store=store, name="case_qc_pass", internal_id="case_qc_pass")
    case_qc_fail = helpers.add_case(store=store, name="case_qc_fail", internal_id="case_qc_fail")

    case_qc_fail_with_failing_controls = helpers.add_case(
        store=store,
        name="case_qc_fail_with_failing_controls",
        internal_id="case_qc_fail_with_failing_controls",
    )

    # Add samples
    sample_qc_pass = helpers.add_sample(
        store=store,
        internal_id="sample_qc_pass",
        name="sample_qc_pass",
        control=ControlOptions.EMPTY,
        reads=861966,
        application_tag=application.tag,
    )

    sample_qc_fail = helpers.add_sample(
        store=store,
        internal_id="sample_qc_fail",
        name="sample_qc_fail",
        control=ControlOptions.EMPTY,
        reads=438776,
        application_tag=application.tag,
    )

    external_negative_control_qc_pass = helpers.add_sample(
        store=store,
        internal_id="external_negative_control_qc_pass",
        name="external_negative_control_qc_pass",
        control=ControlOptions.NEGATIVE,
        reads=20674,
        application_tag=application.tag,
    )

    internal_negative_control_qc_pass = helpers.add_sample(
        store=store,
        internal_id="internal_negative_control_qc_pass",
        name="internal_negative_control_qc_pass",
        control=ControlOptions.NEGATIVE,
        reads=0,
        application_tag=application.tag,
    )

    sample_qc_pass_with_failing_controls = helpers.add_sample(
        store=store,
        internal_id="sample_qc_pass_with_failing_controls",
        name="sample_qc_pass",
        control=ControlOptions.EMPTY,
        reads=861966,
        application_tag=application.tag,
    )

    internal_negative_control_qc_fail = helpers.add_sample(
        store=store,
        internal_id="internal_negative_control_qc_fail",
        name="internal_negative_control_qc_fail",
        control=ControlOptions.NEGATIVE,
        reads=3000,
        application_tag=application.tag,
    )

    external_negative_control_qc_fail = helpers.add_sample(
        store=store,
        internal_id="external_negative_control_qc_fail",
        name="external_negative_control_qc_fail",
        control=ControlOptions.NEGATIVE,
        reads=200000,
        application_tag=application.tag,
    )

    # Add CaseSample relationships
    # case_qc_pass
    helpers.add_relationship(store=store, case=case_qc_pass, sample=sample_qc_pass)
    helpers.add_relationship(
        store=store, case=case_qc_pass, sample=external_negative_control_qc_pass
    )

    # case_qc_fail
    helpers.add_relationship(store=store, case=case_qc_fail, sample=sample_qc_fail)
    helpers.add_relationship(
        store=store, case=case_qc_fail, sample=external_negative_control_qc_pass
    )

    # case_qc_fail_with_failing_controls
    helpers.add_relationship(
        store=store,
        case=case_qc_fail_with_failing_controls,
        sample=sample_qc_pass_with_failing_controls,
    )
    helpers.add_relationship(
        store=store,
        case=case_qc_fail_with_failing_controls,
        sample=external_negative_control_qc_fail,
    )

    return store


@pytest.fixture
def mutant_lims(lims_api: MockLimsAPI) -> MockLimsAPI:
    # Get samples
    sample_qc_pass = LimsSample(id="sample_qc_pass", name="sample_qc_pass")

    sample_qc_fail = LimsSample(id="sample_qc_fail", name="sample_qc_fail")

    external_negative_control_qc_pass = LimsSample(
        id="external_negative_control_qc_pass",
        name="external_negative_control_qc_pass",
        udfs=LimsUDF(control="negative"),
    )
    internal_negative_control_qc_pass = LimsSample(
        id="internal_negative_control_qc_pass",
        name="internal_negative_control_qc_pass",
        udfs=LimsUDF(control="negative", customer="cust000"),
    )

    sample_qc_pass_with_failing_controls = LimsSample(
        id="sample_qc_pass_with_failing_controls", name="sample_qc_pass"
    )

    external_negative_control_qc_fail = LimsSample(
        id="external_negative_control_qc_fail",
        name="external_negative_control_qc_fail",
        udfs=LimsUDF(control="negative"),
    )

    internal_negative_control_qc_fail = LimsSample(
        id="internal_negative_control_qc_fail",
        name="internal_negative_control_qc_fail",
        udfs=LimsUDF(control="negative", customer="cust000"),
    )

    # Create pools
    samples_qc_pass = [
        sample_qc_pass,
        external_negative_control_qc_pass,
        internal_negative_control_qc_pass,
    ]

    samples_qc_fail = [
        sample_qc_fail,
        external_negative_control_qc_pass,
        internal_negative_control_qc_pass,
    ]

    samples_qc_fail_with_failing_controls = [
        sample_qc_pass_with_failing_controls,
        external_negative_control_qc_fail,
        internal_negative_control_qc_fail,
    ]

    # Add pool artifacts
    lims_api.add_artifact_for_sample(sample_id=sample_qc_pass.id, samples=samples_qc_pass)
    lims_api.add_artifact_for_sample(sample_id=sample_qc_fail.id, samples=samples_qc_fail)

    lims_api.add_artifact_for_sample(
        sample_id=sample_qc_pass_with_failing_controls.id,
        samples=samples_qc_fail_with_failing_controls,
    )

    return lims_api


@pytest.fixture
def mutant_quality_controller(
    mutant_store: Store, mutant_lims: MockLimsAPI
) -> MutantQualityController:
    return MutantQualityController(status_db=mutant_store, lims=mutant_lims)


# Samples
@pytest.fixture
def sample_qc_pass(mutant_store: Store) -> Sample:
    return mutant_store.get_sample_by_internal_id("sample_qc_pass")


@pytest.fixture
def internal_negative_control_qc_pass(mutant_store: Store) -> Sample:
    return mutant_store.get_sample_by_internal_id("internal_negative_control_qc_pass")


@pytest.fixture
def external_negative_control_qc_pass(mutant_store: Store) -> Sample:
    return mutant_store.get_sample_by_internal_id("external_negative_control_qc_pass")


@pytest.fixture
def sample_qc_fail(mutant_store: Store) -> Sample:
    return mutant_store.get_sample_by_internal_id("sample_qc_fail")


# Cases
## mutant_case_qc_pass
@pytest.fixture
def mutant_case_qc_pass(mutant_store: Store) -> Case:
    return mutant_store.get_case_by_internal_id("case_qc_pass")


@pytest.fixture
def mutant_analysis_dir_case_qc_pass(mutant_analysis_dir: Path, mutant_case_qc_pass: Case) -> Path:
    return Path(mutant_analysis_dir, mutant_case_qc_pass.internal_id)


@pytest.fixture
def mutant_results_file_path_case_qc_pass(mutant_analysis_dir_case_qc_pass: Path) -> Path:
    return Path(mutant_analysis_dir_case_qc_pass, "pass_sars-cov-2_208455_results.csv")


@pytest.fixture
def mutant_qc_report_path_case_qc_pass(mutant_analysis_dir_case_qc_pass: Path) -> Path:
    return mutant_analysis_dir_case_qc_pass.joinpath(MutantQC.QUALITY_REPORT_FILE_NAME)


@pytest.fixture
def mutant_results_list_qc_pass(mutant_results_file_path_case_qc_pass: Path):
    return _get_validated_results_list(results_file_path=mutant_results_file_path_case_qc_pass)


@pytest.fixture
def mutant_sample_pool_and_results_case_qc_pass(
    mutant_quality_controller: MutantQualityController,
    mutant_results_file_path_case_qc_pass: Path,
    mutant_case_qc_pass: Case,
) -> SamplePoolAndResults:
    return mutant_quality_controller._get_sample_pool_and_results(
        case_results_file_path=mutant_results_file_path_case_qc_pass,
        case=mutant_case_qc_pass,
    )


@pytest.fixture
def mutant_samples_results_case_qc_pass(
    mutant_case_qc_pass: Case, mutant_results_file_path_case_qc_pass: Path
) -> dict[str, ParsedSampleResults]:
    return parse_samples_results(
        case=mutant_case_qc_pass, results_file_path=mutant_results_file_path_case_qc_pass
    )


@pytest.fixture
def mutant_sample_results_sample_qc_pass(
    sample_qc_pass: Sample, mutant_samples_results_case_qc_pass: dict[str, ParsedSampleResults]
) -> ParsedSampleResults:
    sample_results = mutant_samples_results_case_qc_pass[sample_qc_pass.internal_id]
    return sample_results


@pytest.fixture
def mutant_sample_results_external_negative_control_qc_pass(
    external_negative_control_qc_pass: Sample,
    mutant_samples_results_case_qc_pass: dict[str, ParsedSampleResults],
) -> ParsedSampleResults:
    sample_results = mutant_samples_results_case_qc_pass[
        external_negative_control_qc_pass.internal_id
    ]
    return sample_results


@pytest.fixture
def samples_quality_results_case_qc_pass(
    mutant_quality_controller: MutantQualityController,
    mutant_sample_pool_and_results_case_qc_pass: SamplePoolAndResults,
) -> SamplesQualityResults:
    return mutant_quality_controller._get_samples_quality_results(
        sample_pool_and_results=mutant_sample_pool_and_results_case_qc_pass
    )


## mutant_case_qc_fail
@pytest.fixture
def mutant_case_qc_fail(mutant_store: Store) -> Case:
    return mutant_store.get_case_by_internal_id("case_qc_fail")


@pytest.fixture
def mutant_analysis_dir_case_qc_fail(mutant_analysis_dir: Path, mutant_case_qc_fail: Case) -> Path:
    return Path(mutant_analysis_dir, mutant_case_qc_fail.internal_id)


@pytest.fixture
def mutant_results_file_path_qc_fail(mutant_analysis_dir_case_qc_fail: Path) -> Path:
    return Path(mutant_analysis_dir_case_qc_fail, "fail_sars-cov-2_841080_results.csv")


@pytest.fixture
def mutant_qc_report_path_case_qc_fail(mutant_analysis_dir_case_qc_fail: Path) -> Path:
    return mutant_analysis_dir_case_qc_fail.joinpath(MutantQC.QUALITY_REPORT_FILE_NAME)


## mutant_case_qc_fail_with_failing_controls
@pytest.fixture
def mutant_case_qc_fail_with_failing_controls(mutant_store: Store) -> Case:
    return mutant_store.get_case_by_internal_id("case_qc_fail_with_failing_controls")


@pytest.fixture
def mutant_analysis_dir_case_qc_fail_with_failing_controls(
    mutant_analysis_dir: Path, mutant_case_qc_fail_with_failing_controls: Case
) -> Path:
    return Path(mutant_analysis_dir, mutant_case_qc_fail_with_failing_controls.internal_id)


@pytest.fixture
def mutant_results_file_path_qc_fail_with_failing_controls(
    mutant_analysis_dir_case_qc_fail_with_failing_controls: Path,
) -> Path:
    return Path(
        mutant_analysis_dir_case_qc_fail_with_failing_controls,
        "fail_with_failing_controls_sars-cov-2_841080_results.csv",
    )


@pytest.fixture
def mutant_qc_report_path_case_qc_fail_with_failing_controls(
    mutant_analysis_dir_case_qc_fail_with_failing_controls: Path,
) -> Path:
    return mutant_analysis_dir_case_qc_fail_with_failing_controls.joinpath(
        MutantQC.QUALITY_REPORT_FILE_NAME
    )



"""Module for PacBio fixtures returning service objects."""

from pathlib import Path
from unittest.mock import Mock

import pytest

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.services.run_devices.pacbio.data_storage_service.pacbio_store_service import (
    PacBioStoreService,
)
from cg.services.run_devices.pacbio.data_transfer_service.data_transfer_service import (
    PacBioDataTransferService,
)
from cg.services.run_devices.pacbio.housekeeper_service.pacbio_houskeeper_service import (
    PacBioHousekeeperService,
)
from cg.services.run_devices.pacbio.metrics_parser.metrics_parser import PacBioMetricsParser
from cg.services.run_devices.pacbio.post_processing_service import PacBioPostProcessingService
from cg.services.run_devices.pacbio.run_data_generator.pacbio_run_data_generator import (
    PacBioRunDataGenerator,
)
from cg.services.run_devices.pacbio.run_file_manager.run_file_manager import PacBioRunFileManager
from cg.services.run_devices.pacbio.run_validator.pacbio_run_validator import PacBioRunValidator
from cg.services.run_devices.run_names.pacbio import PacbioRunNamesService
from cg.store.store import Store

# Mocked services


@pytest.fixture
def mock_pacbio_run_validator() -> PacBioRunValidator:
    return PacBioRunValidator(
        decompressor=Mock(),
        file_transfer_validator=Mock(),
        file_manager=Mock(),
    )


# Real services


@pytest.fixture
def pac_bio_run_data_generator() -> PacBioRunDataGenerator:
    return PacBioRunDataGenerator()


@pytest.fixture
def pac_bio_run_file_manager() -> PacBioRunFileManager:
    return PacBioRunFileManager()


@pytest.fixture
def pacbio_run_names_service(pac_bio_runs_dir: Path) -> PacbioRunNamesService:
    return PacbioRunNamesService(run_directory=pac_bio_runs_dir.as_posix())


@pytest.fixture
def pac_bio_metrics_parser(pac_bio_run_file_manager: PacBioRunFileManager) -> PacBioMetricsParser:
    return PacBioMetricsParser(file_manager=pac_bio_run_file_manager)


@pytest.fixture
def pac_bio_data_transfer_service(
    pac_bio_metrics_parser: PacBioMetricsParser,
    pac_bio_run_file_manager: PacBioRunFileManager,
) -> PacBioDataTransferService:
    return PacBioDataTransferService(metrics_service=pac_bio_metrics_parser)


@pytest.fixture
def pac_bio_store_service(
    store: Store, pac_bio_data_transfer_service: PacBioDataTransferService
) -> PacBioStoreService:
    return PacBioStoreService(store=store, data_transfer_service=pac_bio_data_transfer_service)


@pytest.fixture
def pac_bio_post_processing_service(
    pac_bio_run_data_generator: PacBioRunDataGenerator,
    pac_bio_housekeeper_service: PacBioHousekeeperService,
    pac_bio_store_service: PacBioStoreService,
    pac_bio_runs_dir: Path,
) -> PacBioPostProcessingService:
    return PacBioPostProcessingService(
        run_validator=Mock(),
        run_data_generator=pac_bio_run_data_generator,
        hk_service=pac_bio_housekeeper_service,
        store_service=pac_bio_store_service,
        sequencing_dir=pac_bio_runs_dir.as_posix(),
    )


@pytest.fixture
def pac_bio_housekeeper_service(
    real_housekeeper_api: HousekeeperAPI,
    pac_bio_run_file_manager: PacBioRunFileManager,
    pac_bio_metrics_parser: PacBioMetricsParser,
) -> PacBioHousekeeperService:
    return PacBioHousekeeperService(
        hk_api=real_housekeeper_api,
        file_manager=pac_bio_run_file_manager,
        metrics_parser=pac_bio_metrics_parser,
    )



import pytest
from sqlalchemy import update

from cg.models.cg_config import CGConfig
from cg.store.store import Store


@pytest.fixture
def nipt_upload_api_context(
    cg_context: CGConfig, re_sequenced_sample_illumina_data_store: Store
) -> CGConfig:
    cg_context.status_db_ = re_sequenced_sample_illumina_data_store
    return cg_context


@pytest.fixture
def nipt_upload_api_failed_fc_context(
    nipt_upload_api_context: CGConfig,
    case_id_for_sample_on_multiple_flow_cells: str,
    re_sequenced_sample_illumina_data_store: Store,
) -> CGConfig:
    nipt_upload_api_failed_fc_context: CGConfig = nipt_upload_api_context
    nipt_upload_api_failed_fc_context.status_db_ = re_sequenced_sample_illumina_data_store
    status_db = nipt_upload_api_context.status_db

    application = status_db.get_application_by_case(case_id_for_sample_on_multiple_flow_cells)
    application.target_reads: int = 1_000_000_000_000

    return nipt_upload_api_failed_fc_context



"""Module for the factory of the deliver files service."""

from typing import Type
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.tb import TrailblazerAPI
from cg.constants import Workflow, DataDelivery
from cg.services.analysis_service.analysis_service import AnalysisService
from cg.services.deliver_files.file_filter.sample_service import SampleFileFilter
from cg.services.deliver_files.tag_fetcher.bam_service import (
    BamDeliveryTagsFetcher,
)
from cg.services.fastq_concatenation_service.fastq_concatenation_service import (
    FastqConcatenationService,
)
from cg.services.deliver_files.deliver_files_service.deliver_files_service import (
    DeliverFilesService,
)
from cg.services.deliver_files.deliver_files_service.exc import DeliveryTypeNotSupported
from cg.services.deliver_files.tag_fetcher.abstract import (
    FetchDeliveryFileTagsService,
)
from cg.services.deliver_files.tag_fetcher.sample_and_case_service import (
    SampleAndCaseDeliveryTagsFetcher,
)
from cg.services.deliver_files.file_fetcher.analysis_service import (
    AnalysisDeliveryFileFetcher,
)
from cg.services.deliver_files.file_fetcher.abstract import (
    FetchDeliveryFilesService,
)
from cg.services.deliver_files.file_fetcher.analysis_raw_data_service import (
    RawDataAndAnalysisDeliveryFileFetcher,
)
from cg.services.deliver_files.file_fetcher.raw_data_service import (
    RawDataDeliveryFileFetcher,
)
from cg.services.deliver_files.file_formatter.service import (
    DeliveryFileFormatter,
)

from cg.services.deliver_files.file_formatter.abstract import (
    DeliveryFileFormattingService,
)
from cg.services.deliver_files.file_formatter.utils.case_service import (
    CaseFileFormatter,
)
from cg.services.deliver_files.file_formatter.utils.sample_concatenation_service import (
    SampleFileConcatenationFormatter,
)
from cg.services.deliver_files.file_formatter.utils.sample_service import (
    SampleFileFormatter,
)
from cg.services.deliver_files.file_mover.service import (
    DeliveryFilesMover,
)
from cg.services.deliver_files.rsync.service import (
    DeliveryRsyncService,
)
from cg.store.store import Store


class DeliveryServiceFactory:
    """Class to build the delivery services based on workflow and delivery type."""

    def __init__(
        self,
        store: Store,
        hk_api: HousekeeperAPI,
        rsync_service: DeliveryRsyncService,
        tb_service: TrailblazerAPI,
        analysis_service: AnalysisService,
    ):
        self.store = store
        self.hk_api = hk_api
        self.rsync_service = rsync_service
        self.tb_service = tb_service
        self.analysis_service = analysis_service

    @staticmethod
    def _get_file_tag_fetcher(delivery_type: DataDelivery) -> FetchDeliveryFileTagsService:
        """Get the file tag fetcher based on the delivery type."""
        service_map: dict[DataDelivery, Type[FetchDeliveryFileTagsService]] = {
            DataDelivery.FASTQ: SampleAndCaseDeliveryTagsFetcher,
            DataDelivery.ANALYSIS_FILES: SampleAndCaseDeliveryTagsFetcher,
            DataDelivery.FASTQ_ANALYSIS: SampleAndCaseDeliveryTagsFetcher,
            DataDelivery.BAM: BamDeliveryTagsFetcher,
        }
        return service_map[delivery_type]()

    def _get_file_fetcher(self, delivery_type: DataDelivery) -> FetchDeliveryFilesService:
        """Get the file fetcher based on the delivery type."""
        service_map: dict[DataDelivery, Type[FetchDeliveryFilesService]] = {
            DataDelivery.FASTQ: RawDataDeliveryFileFetcher,
            DataDelivery.ANALYSIS_FILES: AnalysisDeliveryFileFetcher,
            DataDelivery.FASTQ_ANALYSIS: RawDataAndAnalysisDeliveryFileFetcher,
            DataDelivery.BAM: RawDataDeliveryFileFetcher,
        }
        file_tag_fetcher: FetchDeliveryFileTagsService = self._get_file_tag_fetcher(delivery_type)
        return service_map[delivery_type](
            status_db=self.store,
            hk_api=self.hk_api,
            tags_fetcher=file_tag_fetcher,
        )

    @staticmethod
    def _get_sample_file_formatter(
        workflow: Workflow,
    ) -> SampleFileFormatter | SampleFileConcatenationFormatter:
        """Get the file formatter service based on the workflow."""
        if workflow in [Workflow.MICROSALT]:
            return SampleFileConcatenationFormatter(FastqConcatenationService())
        return SampleFileFormatter()

    @staticmethod
    def _validate_delivery_type(delivery_type: DataDelivery):
        """Check if the delivery type is supported. Raises DeliveryTypeNotSupported error."""
        if delivery_type in [
            DataDelivery.FASTQ,
            DataDelivery.ANALYSIS_FILES,
            DataDelivery.FASTQ_ANALYSIS,
            DataDelivery.BAM,
        ]:
            return
        raise DeliveryTypeNotSupported(
            f"Delivery type {delivery_type} is not supported. Supported delivery types are {DataDelivery.FASTQ}, {DataDelivery.ANALYSIS_FILES}, {DataDelivery.FASTQ_ANALYSIS}, {DataDelivery.BAM}."
        )

    def build_delivery_service(
        self, workflow: Workflow, delivery_type: DataDelivery
    ) -> DeliverFilesService:
        """Build a delivery service based on the workflow and delivery type."""
        delivery_type: DataDelivery = self._sanitise_delivery_type(delivery_type)
        self._validate_delivery_type(delivery_type)
        file_fetcher: FetchDeliveryFilesService = self._get_file_fetcher(delivery_type)
        sample_file_formatter: SampleFileFormatter | SampleFileConcatenationFormatter = (
            self._get_sample_file_formatter(workflow)
        )
        file_formatter: DeliveryFileFormattingService = DeliveryFileFormatter(
            case_file_formatter=CaseFileFormatter(), sample_file_formatter=sample_file_formatter
        )
        return DeliverFilesService(
            delivery_file_manager_service=file_fetcher,
            move_file_service=DeliveryFilesMover(),
            file_filter=SampleFileFilter(),
            file_formatter_service=file_formatter,
            status_db=self.store,
            rsync_service=self.rsync_service,
            tb_service=self.tb_service,
            analysis_service=self.analysis_service,
        )

    @staticmethod
    def _sanitise_delivery_type(delivery_type: DataDelivery) -> DataDelivery:
        """Sanitise the delivery type."""
        if delivery_type in [DataDelivery.FASTQ_QC, DataDelivery.FASTQ_SCOUT]:
            return DataDelivery.FASTQ
        if delivery_type in [DataDelivery.ANALYSIS_SCOUT]:
            return DataDelivery.ANALYSIS_FILES
        if delivery_type in [
            DataDelivery.FASTQ_ANALYSIS_SCOUT,
            DataDelivery.FASTQ_QC_ANALYSIS,
        ]:
            return DataDelivery.FASTQ_ANALYSIS
        return delivery_type
